# encoding: utf-8
#
#Copyright (C) 2017-2023, P. R. Wiecha
#
#This program is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
python implementations of the Green's Dyads, accelerated using `numba`
"""
from __future__ import print_function
from __future__ import absolute_import

import copy
import time
import warnings

import multiprocessing
import math
import cmath
import numpy as np

import numba

from pyGDM2.propagators import propagators



@numba.njit(cache=True)
def moving_average(a, n=15):
    ret = np.cumsum(a)
    ret[n:] = ret[n:] - ret[:-n]
    return ret[n - 1:] / n


### --- 3D periodic Dyads - until now homogeneous environment only, normal incidence only
@numba.njit(cache=True)
def G0_periodic_123_direct_sum(R1, R2, wavelength, 
                    eps1=1.0+0j, eps2=1.0+0j, eps3=1.0+0j, spacing=5000, 
                    k_x=0.0, N_u=0, N_v=0, 
                    u=np.array([0., 0., 0.]), v=np.array([0., 0., 0.])):
    xx = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    yy = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    zz = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    xy = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    xz = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    yx = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    yz = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    zx = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    zy = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex64)
    
    for u_i in numba.prange(-N_u, N_u+1):
        for v_i in numba.prange(-N_v, N_v+1):
            _R1 = R1 + u*u_i + v*v_i     # position of emitter-copy from periodic cell [u_i, v_i]
            _xx, _yy, _zz, _xy, _xz, _yx, _yz, _zx, _zy = propagators.G0_EE_123(
                    _R1, R2, wavelength, eps1, eps2, eps3, spacing)
            
            ## !!! exp_term_k_x = np.exp(1j* k_x * (u*u_i + v*v_i)[0])  # should be added, see HE-tensor
            xx[u_i, v_i] = _xx
            yy[u_i, v_i] = _yy
            zz[u_i, v_i] = _zz
            xy[u_i, v_i] = _xy
            xz[u_i, v_i] = _xz
            yx[u_i, v_i] = _yx
            yz[u_i, v_i] = _yz
            zx[u_i, v_i] = _zx
            zy[u_i, v_i] = _zy
    
    return np.sum(xx), np.sum(yy), np.sum(zz), \
            np.sum(xy), np.sum(xz), np.sum(yx), \
            np.sum(yz), np.sum(zx), np.sum(zy)




### --- 3D periodic Dyads - until now homogeneous environment only, normal incidence only
@numba.njit(cache=True)
def G0_periodic_123(R1, R2, wavelength, 
                    eps1=1.0+0j, eps2=1.0+0j, eps3=1.0+0j, spacing=5000, 
                    k_x=0.0, N_u=0, N_v=0, 
                    u=np.array([0., 0., 0.]), v=np.array([0., 0., 0.])):
    xx = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)   # np.cumsum suffers from roudoff errors --> 128bit complex
    yy = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
    zz = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
    xy = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
    xz = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
    yx = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
    yz = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
    zx = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
    zy = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
    
    for _u_i in numba.prange(1, 2*(N_u+1)):
        u_i = (_u_i%2-0.5)*2 * _u_i//2       # start at center then hopping increasingly +/-1 in both directions
        for _v_i in numba.prange(1, 2*(N_v+1)):
            v_i = (_v_i%2-0.5)*2 * _v_i//2 
            _R1 = R1 + u*u_i + v*v_i     # position of emitter-copy from periodic cell [u_i, v_i]
            _xx, _yy, _zz, _xy, _xz, _yx, _yz, _zx, _zy = propagators.G0_EE_123(
                    _R1, R2, wavelength, eps1, eps2, eps3, spacing)
            
            ## !!! exp_term_k_x = np.exp(1j* k_x * (u*u_i + v*v_i)[0])  # should be added, see HE-tensor
            xx[_u_i-1, _v_i-1] = _xx
            yy[_u_i-1, _v_i-1] = _yy
            zz[_u_i-1, _v_i-1] = _zz
            xy[_u_i-1, _v_i-1] = _xy
            xz[_u_i-1, _v_i-1] = _xz
            yx[_u_i-1, _v_i-1] = _yx
            yz[_u_i-1, _v_i-1] = _yz
            zx[_u_i-1, _v_i-1] = _zx
            zy[_u_i-1, _v_i-1] = _zy
    
    # return np.sum(xx), np.sum(yy), np.sum(zz), \
    #        np.sum(xy), np.sum(xz), np.sum(yx), \
    #        np.sum(yz), np.sum(zx), np.sum(zy)
    
    ## calc. mean of moving average of tail of series to dampen oscillations
    xx = np.mean(moving_average(np.cumsum(xx.flatten()), n=N_u+N_v)[2*N_u*N_v:-1])
    yy = np.mean(moving_average(np.cumsum(yy.flatten()), n=N_u+N_v)[2*N_u*N_v:-1])
    zz = np.mean(moving_average(np.cumsum(zz.flatten()), n=N_u+N_v)[2*N_u*N_v:-1])
    xy = np.mean(moving_average(np.cumsum(xy.flatten()), n=N_u+N_v)[2*N_u*N_v:-1])
    xz = np.mean(moving_average(np.cumsum(xz.flatten()), n=N_u+N_v)[2*N_u*N_v:-1])
    yx = np.mean(moving_average(np.cumsum(yx.flatten()), n=N_u+N_v)[2*N_u*N_v:-1])
    yz = np.mean(moving_average(np.cumsum(yz.flatten()), n=N_u+N_v)[2*N_u*N_v:-1])
    zx = np.mean(moving_average(np.cumsum(zx.flatten()), n=N_u+N_v)[2*N_u*N_v:-1])
    zy = np.mean(moving_average(np.cumsum(zy.flatten()), n=N_u+N_v)[2*N_u*N_v:-1])
    
    return xx, yy, zz, xy, xz, yx, yz, zx, zy


@numba.njit(cache=True)
def Gs_periodic_123(R1, R2, wavelength, 
                    eps1=1.0+0j, eps2=1.0+0j, eps3=1.0+0j, spacing=5000, 
                    k_x=0.0, N_u=0, N_v=0, 
                    u=np.array([0., 0., 0.]), v=np.array([0., 0., 0.])):
    xx = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)   # np.cumsum suffers from roudoff errors --> 128bit complex
    yy = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
    zz = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
    xy = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
    xz = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
    yx = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
    yz = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
    zx = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
    zy = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
    
    for _u_i in numba.prange(1, 2*(N_u+1)):
        u_i = (_u_i%2-0.5)*2 * _u_i//2
        for _v_i in numba.prange(1, 2*(N_v+1)):
            v_i = (_v_i%2-0.5)*2 * _v_i//2 
            _R1 = R1 + u*u_i + v*v_i     # position of emitter-copy from periodic cell [u_i, v_i]
            _xx, _yy, _zz, _xy, _xz, _yx, _yz, _zx, _zy = propagators.Gs_EE_123(
                    _R1, R2, wavelength, eps1, eps2, eps3, spacing)
            
            ## !!! exp_term_k_x = np.exp(1j* k_x * (u*u_i + v*v_i)[0])  # should be added, see HE-tensor
            xx[_u_i-1, _v_i-1] = _xx
            yy[_u_i-1, _v_i-1] = _yy
            zz[_u_i-1, _v_i-1] = _zz
            xy[_u_i-1, _v_i-1] = _xy
            xz[_u_i-1, _v_i-1] = _xz
            yx[_u_i-1, _v_i-1] = _yx
            yz[_u_i-1, _v_i-1] = _yz
            zx[_u_i-1, _v_i-1] = _zx
            zy[_u_i-1, _v_i-1] = _zy
            
    # return np.sum(xx), np.sum(yy), np.sum(zz), \
    #        np.sum(xy), np.sum(xz), np.sum(yx), \
    #        np.sum(yz), np.sum(zx), np.sum(zy)
    
    ## calc. mean of moving average of tail of series to dampen oscillations
    xx = np.mean(moving_average(np.cumsum(xx.flatten()), n=N_u*N_v)[2*N_u*N_v:-1])
    yy = np.mean(moving_average(np.cumsum(yy.flatten()), n=N_u*N_v)[2*N_u*N_v:-1])
    zz = np.mean(moving_average(np.cumsum(zz.flatten()), n=N_u*N_v)[2*N_u*N_v:-1])
    xy = np.mean(moving_average(np.cumsum(xy.flatten()), n=N_u*N_v)[2*N_u*N_v:-1])
    xz = np.mean(moving_average(np.cumsum(xz.flatten()), n=N_u*N_v)[2*N_u*N_v:-1])
    yx = np.mean(moving_average(np.cumsum(yx.flatten()), n=N_u*N_v)[2*N_u*N_v:-1])
    yz = np.mean(moving_average(np.cumsum(yz.flatten()), n=N_u*N_v)[2*N_u*N_v:-1])
    zx = np.mean(moving_average(np.cumsum(zx.flatten()), n=N_u*N_v)[2*N_u*N_v:-1])
    zy = np.mean(moving_average(np.cumsum(zy.flatten()), n=N_u*N_v)[2*N_u*N_v:-1])
    
    return xx, yy, zz, xy, xz, yx, yz, zx, zy


@numba.njit(cache=True)
def Gtot_periodic_123(R1, R2, wavelength, 
                      eps1=1.0+0j, eps2=1.0+0j, eps3=1.0+0j, spacing=5000, 
                      k_x=0.0, N_u=0, N_v=0, 
                      u=np.array([0., 0., 0.]), v=np.array([0., 0., 0.])):
    ## ----- free space term
    xx, yy, zz, xy, xz, yx, yz, zx, zy = G0_periodic_123(R1, R2, wavelength, 
                                eps1, eps2, eps3, spacing, k_x, N_u, N_v, u, v)
    
    ## ----- surface term
    xxs,yys,zzs,xys,xzs,yxs,yzs,zxs,zys = Gs_periodic_123(R1, R2, wavelength, 
                                eps1, eps2, eps3, spacing, k_x, N_u, N_v, u, v)

    return xx + xxs, yy + yys, zz + zzs, \
           xy + xys, xz + xzs, yx + yxs, \
           yz + yzs, zx + zxs, zy + zys


@numba.njit(cache=True)
def G_HE_periodic(R1, R2, wavelength, 
                    eps1=1.0+0j, eps2=1.0+0j, eps3=1.0+0j, spacing=5000, 
                    k_x=0.0, N_u=0, N_v=0, 
                    u=np.array([0., 0., 0.]), v=np.array([0., 0., 0.])):
    xx = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)   # np.cumsum suffers from roudoff errors --> 128bit complex
    yy = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
    zz = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
    xy = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
    xz = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
    yx = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
    yz = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
    zx = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
    zy = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
    
        
    for _u_i in numba.prange(1, 2*(N_u+1)):
        u_i = (_u_i%2-0.5)*2 * _u_i//2
        for _v_i in numba.prange(1, 2*(N_v+1)):
            v_i = (_v_i%2-0.5)*2 * _v_i//2 
            _R1 = R1 + u*u_i + v*v_i     # position of emitter-copy from periodic cell [u_i, v_i]
            _xx, _yy, _zz, _xy, _xz, _yx, _yz, _zx, _zy = propagators.G_HE_123(
                    _R1, R2, wavelength, eps1, eps2, eps3, spacing)
            
            ## additional phase, use x-component of wave-vector --> incidence in XZ plane!
            exp_term_k_x = np.exp(1j* k_x * (u*u_i + v*v_i)[0])
            xx[_u_i-1, _v_i-1] = _xx * exp_term_k_x
            yy[_u_i-1, _v_i-1] = _yy * exp_term_k_x
            zz[_u_i-1, _v_i-1] = _zz * exp_term_k_x
            xy[_u_i-1, _v_i-1] = _xy * exp_term_k_x
            xz[_u_i-1, _v_i-1] = _xz * exp_term_k_x
            yx[_u_i-1, _v_i-1] = _yx * exp_term_k_x
            yz[_u_i-1, _v_i-1] = _yz * exp_term_k_x
            zx[_u_i-1, _v_i-1] = _zx * exp_term_k_x
            zy[_u_i-1, _v_i-1] = _zy * exp_term_k_x
    
    # return np.sum(xx), np.sum(yy), np.sum(zz), \
    #        np.sum(xy), np.sum(xz), np.sum(yx), \
    #        np.sum(yz), np.sum(zx), np.sum(zy)
    
    ## calc. mean of moving average of tail of series to dampen oscillations
    xx = np.mean(moving_average(np.cumsum(xx.flatten()), n=N_u*N_v)[2*N_u*N_v:-1])
    yy = np.mean(moving_average(np.cumsum(yy.flatten()), n=N_u*N_v)[2*N_u*N_v:-1])
    zz = np.mean(moving_average(np.cumsum(zz.flatten()), n=N_u*N_v)[2*N_u*N_v:-1])
    xy = np.mean(moving_average(np.cumsum(xy.flatten()), n=N_u*N_v)[2*N_u*N_v:-1])
    xz = np.mean(moving_average(np.cumsum(xz.flatten()), n=N_u*N_v)[2*N_u*N_v:-1])
    yx = np.mean(moving_average(np.cumsum(yx.flatten()), n=N_u*N_v)[2*N_u*N_v:-1])
    yz = np.mean(moving_average(np.cumsum(yz.flatten()), n=N_u*N_v)[2*N_u*N_v:-1])
    zx = np.mean(moving_average(np.cumsum(zx.flatten()), n=N_u*N_v)[2*N_u*N_v:-1])
    zy = np.mean(moving_average(np.cumsum(zy.flatten()), n=N_u*N_v)[2*N_u*N_v:-1])
    
    return xx, yy, zz, xy, xz, yx, yz, zx, zy
           


# @numba.njit(cache=True)
# def G_EE_perdiodic(R1, R2, wavelength, conf_dict):
#     """evaluate field of dipole emitter in 1-2-3 layer environment (NF approx.)
    
#     R1: dipole position
#     R2: evaluation position
#     """
#     k_x = np.float32(conf_dict['k_x'].real)
#     # eps1 = conf_dict['eps1']
#     eps2 = conf_dict['eps2']
#     # eps3 = conf_dict['eps3']
#     # spacing = np.float32(conf_dict['spacing'].real)
#     cutoff_u = np.int(conf_dict['cutoff_u'].real)
#     cutoff_v = np.int(conf_dict['cutoff_v'].real)
#     u = np.array([np.float32(conf_dict['ux'].real), 
#                   np.float32(conf_dict['uy'].real), 
#                   np.float32(conf_dict['uz'].real)])
#     v = np.array([np.float32(conf_dict['vx'].real), 
#                   np.float32(conf_dict['vy'].real), 
#                   np.float32(conf_dict['vz'].real)])
    
#     return _G0_periodic(R1, R2, wavelength, eps2, k_x, cutoff_u, cutoff_v, u, v)


# @numba.njit(cache=True)
# def G_HE_perdiodic(R1, R2, wavelength, conf_dict):
#     """evaluate field of dipole emitter in 1-2-3 layer environment (NF approx.)
    
#     R1: dipole position
#     R2: evaluation position
#     """
#     k_x = np.float32(conf_dict['k_x'].real)
#     # eps1 = conf_dict['eps1']
#     eps2 = conf_dict['eps2']
#     # eps3 = conf_dict['eps3']
#     # spacing = np.float32(conf_dict['spacing'].real)
#     cutoff_u = np.int(conf_dict['cutoff_u'].real)
#     cutoff_v = np.int(conf_dict['cutoff_v'].real)
#     u = np.array([np.float32(conf_dict['ux'].real), 
#                   np.float32(conf_dict['uy'].real), 
#                   np.float32(conf_dict['uz'].real)])
#     v = np.array([np.float32(conf_dict['vx'].real), 
#                   np.float32(conf_dict['vy'].real), 
#                   np.float32(conf_dict['vz'].real)])
    
#     return _G0_HE_periodic(R1, R2, wavelength, eps2, k_x, cutoff_u, cutoff_v, u, v)



### --- coupling matrix gen. non-retarded 
# =============================================================================
# multi-dp-propagation / coupling matrix generators non-retarded 3D
# =============================================================================
## --- multi-dipole field propagation
@numba.njit(parallel=True)
def greens_tensor_evaluation_periodic(dp_pos, r_probe, G_func, wavelength, conf_dict, M,
                                      selfterm=None, dist_div_G=None):
    k_x = np.float32(conf_dict['k_x'].real)
    
    eps1 = conf_dict['eps1']
    eps2 = conf_dict['eps2']
    eps3 = conf_dict['eps3']
    spacing = np.float32(conf_dict['spacing'].real)
    
    cutoff_u = np.int(conf_dict['cutoff_u'].real)
    cutoff_v = np.int(conf_dict['cutoff_v'].real)
    u = np.array([np.float32(conf_dict['ux'].real), 
                  np.float32(conf_dict['uy'].real), 
                  np.float32(conf_dict['uz'].real)])
    v = np.array([np.float32(conf_dict['vx'].real), 
                  np.float32(conf_dict['vy'].real), 
                  np.float32(conf_dict['vz'].real)])
    
    for i in numba.prange(len(dp_pos)):   # explicit parallel loop
        _pos = dp_pos[i]
        for j in range(len(r_probe)):
            _r = r_probe[j]
            xx, yy, zz, xy, xz, yx, yz, zx, zy = G_func(
                                        _pos, _r, wavelength, eps1, eps2, eps3, 
                                        spacing, k_x, cutoff_u, cutoff_v, u, v
                                        )
            ## return list of Greens tensors
            M[i,j,0,0], M[i,j,1,1], M[i,j,2,2] = xx, yy, zz
            M[i,j,1,0], M[i,j,2,0], M[i,j,0,1] = yx, zx, xy
            M[i,j,2,1], M[i,j,0,2], M[i,j,1,2] = zy, xz, yz

#%%
# =============================================================================
# Green's tensor evaluation functions
# =============================================================================
def _get_nr_processes():
    """return available processes
    
    see: 
    https://stackoverflow.com/questions/1006289/how-to-find-out-the-number-of-cpus-using-python
    """
    ## preffered method to get available processes (might fail on windows)
    try:
        import os
        return len(os.sched_getaffinity(0))
    except:
        pass
    
    ## if failed, try alternative using psutils
    try:
        import psutil
        return len(psutil.Process().cpu_affinity())
    except:
        pass
    
    ## fall back on multiprocessing value (if psutils not installed)
    import multiprocessing
    return multiprocessing.cpu_count()


@numba.njit(parallel=True)
def _eval_geo_positions_G0(geo, compare_array):
    for i in numba.prange(len(geo)):    # explicit parallel loop
        R2 = geo[i]       # "observer"
        for j in range(len(geo)):
            R1 = geo[j]   # emitter
            
            DX = R2[0] - R1[0]
            DY = R2[1] - R1[1]
            DZ = R2[2] - R1[2]
            
            compare_array[i,j] = [DX, DY, DZ, DZ]  # last DZ: dummy placeholder
    
@numba.njit(parallel=True)
def _eval_geo_positions_Gs(geo, compare_array):
    for i in numba.prange(len(geo)):    # explicit parallel loop
        R2 = geo[i]       # "observer"
        for j in range(len(geo)):
            R1 = geo[j]   # emitter
            
            DX = R2[0] - R1[0]
            DY = R2[1] - R1[1]
            DZsum = R2[2] + R1[2]
            
            compare_array[i,j] = [DX, DY, DZsum, DZsum]  # last DZsum: dummy placeholder

@numba.njit
def _reconstruct_tensors(M, tensor, index_grp):
    for j in index_grp:
        M[j] = tensor

def _integrate_g0_EE_periodic(args):
    DX,DY,DZ,_, wavelength, eps1, eps2, eps3, \
        spacing, k_x, cutoff_u, cutoff_v, ux,uy,uz, vx,vy,vz = args

    u = np.array([ux.real,uy.real,uz.real]).astype(np.float32)
    v = np.array([vx.real,vy.real,vz.real]).astype(np.float32)
    R1 = np.array([0,0,0], dtype=np.float32)
    R2 = np.array([DX.real, DY.real, DZ.real], dtype=np.float32)
    wavelength = wavelength.real
    k_x = k_x.real
    spacing = spacing.real
    cutoff_u = int(cutoff_u.real)
    cutoff_v = int(cutoff_v.real)
    
    if np.linalg.norm(R2) == 0:  # r=r' --> return 0
        xx, yy, zz, xy, xz, yx, yz, zx, zy = np.zeros(9, dtype=np.complex64)
    else:
        xx, yy, zz, xy, xz, yx, yz, zx, zy = G0_periodic_123(
                            R1, R2, wavelength, eps1, eps2, eps3, 
                            spacing, k_x, cutoff_u, cutoff_v, u, v
                            )
    SEE = np.array([[xx,xy,xz], [yx,yy,yz], [zx,zy,zz]], dtype=np.complex64)
    
    return SEE

def _integrate_gs_EE_periodic(args):
    DX,DY,DZ,_, wavelength, eps1, eps2, eps3, \
        spacing, k_x, cutoff_u, cutoff_v, ux,uy,uz, vx,vy,vz = args
        
    u = np.array([ux.real,uy.real,uz.real]).astype(np.float32)
    v = np.array([vx.real,vy.real,vz.real]).astype(np.float32)
    R1 = np.array([0,0,0], dtype=np.float32)
    R2 = np.array([DX.real, DY.real, DZ.real], dtype=np.float32)
    wavelength = wavelength.real
    k_x = k_x.real
    spacing = spacing.real
    cutoff_u = int(cutoff_u.real)
    cutoff_v = int(cutoff_v.real)
    
    if np.linalg.norm(R2) == 0:
        xx, yy, zz, xy, xz, yx, yz, zx, zy = np.zeros(9, dtype=np.complex64)
    else:
        xx, yy, zz, xy, xz, yx, yz, zx, zy = Gs_periodic_123(
                            R1, R2, wavelength, eps1, eps2, eps3, 
                            spacing, k_x, cutoff_u, cutoff_v, u, v
                            )
    SEEs = np.array([[xx,xy,xz], [yx,yy,yz], [zx,zy,zz]], dtype=np.complex64)
    
    return SEEs


def _integrate_g0_HE_periodic(args):
    DX,DY,DZ,_, wavelength, eps1, eps2, eps3, \
        spacing, k_x, cutoff_u, cutoff_v, ux,uy,uz, vx,vy,vz = args
        
    u = np.array([ux.real,uy.real,uz.real]).astype(np.float32)
    v = np.array([vx.real,vy.real,vz.real]).astype(np.float32)
    R1 = np.array([0,0,0], dtype=np.float32)
    R2 = np.array([DX.real, DY.real, DZ.real], dtype=np.float32)
    wavelength = wavelength.real
    k_x = k_x.real
    spacing = spacing.real
    cutoff_u = int(cutoff_u.real)
    cutoff_v = int(cutoff_v.real)
    
    if np.linalg.norm(R2) == 0:
        xx, yy, zz, xy, xz, yx, yz, zx, zy = np.zeros(9, dtype=np.complex64)
    else:
        xx, yy, zz, xy, xz, yx, yz, zx, zy = G_HE_periodic(
                            R1, R2, wavelength, eps1, eps2, eps3, 
                            spacing, k_x, cutoff_u, cutoff_v, u, v
                            )
    
    SHE = np.array([[xx,xy,xz], [yx,yy,yz], [zx,zy,zz]], dtype=np.complex64)
    return SHE





def t_sbs_G0_periodic(geo, wavelength, conf_dict, 
                      func_eval_identical_tensors=_eval_geo_positions_G0,
                      func_integrate_gs=_integrate_g0_EE_periodic,
                      verbose=True):
    
    eps1 = conf_dict['eps1']
    eps2 = conf_dict['eps2']
    eps3 = conf_dict['eps3']
    spacing = np.float32(conf_dict['spacing'].real)
    
    cutoff_u = int(conf_dict['cutoff_u'].real)
    cutoff_v = int(conf_dict['cutoff_v'].real)
    u = np.array([np.float32(conf_dict['ux'].real), 
                  np.float32(conf_dict['uy'].real), 
                  np.float32(conf_dict['uz'].real)])
    v = np.array([np.float32(conf_dict['vx'].real), 
                  np.float32(conf_dict['vy'].real), 
                  np.float32(conf_dict['vz'].real)])
    
    k_x = np.float32(conf_dict['k_x'].real)
    
    
    ## --- determine unique tensors in coupling matrix
    t0 = time.time()
    compare_array = np.zeros([len(geo), len(geo), 4], dtype=np.float32)
    func_eval_identical_tensors(geo, compare_array)
    compare_array = np.reshape(compare_array, (-1,4))
    if verbose: 
        t1 = time.time()
        print('identification of identical tensors: {:.2f}s'.format(t1-t0))
    
    
    ## --- sort indices by identical tensors
    idx_sort = np.argsort(compare_array.view('f4,f4,f4,f4'), axis=0, order=['f0', 'f1', 'f2'])
    sorted_compare_array = compare_array[idx_sort]
    if verbose: 
        t2 = time.time()
        print('sorting:                             {:.2f}s'.format(t2-t1))
    
        
    ## --- indices of identical tensors in coupling matrix
    unique_arr, idx_start, count = np.unique(sorted_compare_array, axis=0, 
                                             return_counts=True, return_index=True)
    
    ## splits the indices of identical tensors into separate arrays
    index_groups = np.split(idx_sort, idx_start[1:])
    if verbose: 
        t3 = time.time()
        print('indexing of unique tensors:          {:.2f}s'.format(t3-t2))
    
    
    ## --- construct coupling matrix
    ## calculate tensors with multiprocessing
    t0 = time.time()
    ## number of parallel processes
    if int(conf_dict['n_cpu'].real)==-1:
        N_cpu = _get_nr_processes()
    else:
        N_cpu = int(conf_dict['n_cpu'].real)
    
    args = np.ones((len(unique_arr), 14), dtype=np.complex64) * \
           np.array([wavelength, eps1, eps2, eps3, \
                     spacing, k_x, cutoff_u, cutoff_v, *u, *v], dtype=np.complex64)
    argslist = np.concatenate([np.squeeze(unique_arr), args], axis=1)
    
    with multiprocessing.Pool(N_cpu) as p:
        tensor_arr = p.map(func_integrate_gs, argslist)
    if verbose:
        t4 = time.time()
        print('calculating {: >6} periodic tensors: {:.2f}s (working on {} processes)'.format(len(unique_arr), t4-t3, N_cpu))
    
    # ## for testing: non-multiporcessing, sequential
    # tensor_arr = []
    # for arg in argslist:
    #     _G = func_integrate_gs(arg)
    #     tensor_arr.append(_G)
    
    ## --- fill coupling matrix
    M = np.zeros((len(geo)*len(geo), 3, 3), dtype=np.complex64)
    
    for i_G, idx_unique in enumerate(index_groups):
        _reconstruct_tensors(M, tensor_arr[i_G], idx_unique[:,0])
    M = np.reshape(M, [len(geo), len(geo), 3, 3]).swapaxes(0, 1)
    if verbose: 
        t5 = time.time()
        print('reconstruction of coupling matrix:   {:.2f}s'.format(t5-t4))
    
    return M



# @numba.njit(parallel=True, cache=True)
def t_sbs_selfterms_only(geo, wavelength, selfterms, conf_dict, M):
    for i in numba.prange(len(geo)):    # explicit parallel loop
        for j in range(len(geo)):
            if i==j:
                ## self term
                M[i,j] = selfterms[j]


@numba.njit(parallel=True, cache=True)
def _construct_coupled_dipole_system_EE(M0, Mself, Ms, alpha, M_sbs):
    for i in numba.prange(len(M0)):    # explicit parallel loop
        for j in range(len(M0[0])):
            ## invertible matrix:  delta_ij*1 - G[i,j] * alpha[j]
            # M_sbs[3*i:3*i+3, 3*j:3*j+3] = Ms[i,j]    # testing
            _M = M0[i,j] + Mself[i,j] + Ms[i,j]
            M_sbs[3*i:3*i+3, 3*j:3*j+3] = -1 * np.dot(_M, alpha[j])
            if i==j:
                M_sbs[3*i:3*i+3, 3*j:3*j+3] += np.identity(3)


@numba.njit(parallel=True, cache=True)
def _construct_coupled_dipole_system_HE(M0, Mself, Ms, alpha, M_sbs):
    for i in numba.prange(len(M0)):    # explicit parallel loop
        for j in range(len(M0[0])):
            ## invertible matrix:  - G[i,j] * alpha[j]
            _M = M0[i,j] + Mself[i,j] + Ms[i,j]
            M_sbs[3*i:3*i+3, 3*j:3*j+3] = -1 * np.dot(_M, alpha[j])


## !!!!!!!!!!!! MERGE THE FOLLOWING INTO PERIODIC CODE
# =============================================================================
# the actual coupling matrix generator functions
# =============================================================================
def t_sbs_EE_123_quasistatic_periodic(geo, wavelength, selfterms, alpha, conf_dict, M):
    M0 = np.zeros((len(geo), len(geo), 3, 3), dtype=np.complex64)
    Ms = np.zeros((len(geo), len(geo), 3, 3), dtype=np.complex64)
    Mself = np.zeros((len(geo), len(geo), 3, 3), dtype=np.complex64)
    Msbs = np.zeros((len(geo)*3, len(geo)*3), dtype=np.complex64)

    
    M0 = t_sbs_G0_periodic(geo, wavelength, conf_dict, 
                           func_eval_identical_tensors=_eval_geo_positions_G0,
                           func_integrate_gs=_integrate_g0_EE_periodic,
                           verbose=True)
    if conf_dict['eps1'] != conf_dict['eps2'] or conf_dict['eps2'] != conf_dict['eps3']:
        Ms = t_sbs_G0_periodic(geo, wavelength, conf_dict, 
                                func_eval_identical_tensors=_eval_geo_positions_Gs,
                                func_integrate_gs=_integrate_g0_EE_periodic,
                                verbose=True)
    t_sbs_selfterms_only(geo, wavelength, selfterms, conf_dict, Mself)
    
    _construct_coupled_dipole_system_EE(M0, Mself, Ms, alpha, Msbs)
    M[...] = Msbs    # subsitute M in-place
    

def t_sbs_HE_123_quasistatic_periodic(geo, wavelength, selfterms, alpha, conf_dict, M):
    M0 = np.zeros((len(geo), len(geo), 3, 3), dtype=np.complex64)
    Mself = np.zeros((len(geo), len(geo), 3, 3), dtype=np.complex64)
    Msbs = np.zeros((len(geo)*3, len(geo)*3), dtype=np.complex64)
    
    ## quasistatic approx.: surface contribution to G_HE = 0 
    Ms = np.zeros((len(geo), len(geo), 3, 3), dtype=np.complex64)
    
    M0 = t_sbs_G0_periodic(geo, wavelength, conf_dict, 
                           func_eval_identical_tensors=_eval_geo_positions_G0,
                           func_integrate_gs=_integrate_g0_EE_periodic,
                           verbose=True)
    t_sbs_selfterms_only(geo, wavelength, selfterms, conf_dict, Mself)
    
    _construct_coupled_dipole_system_HE(M0, Mself, Ms, alpha, Msbs)
    M[...] = Msbs    # subsitute M in-place
    
    





# #%%
# ## !!!!!! OLD, 'slow' PART
# @numba.njit(parallel=True, cache=True)
# def t_sbs_EE_123_quasistatic_periodic(geo, wavelength, selfterms, alpha, conf_dict, M):
#     k_x = np.float32(conf_dict['k_x'].real)
    
#     eps1 = conf_dict['eps1']
#     eps2 = conf_dict['eps2']
#     eps3 = conf_dict['eps3']
#     spacing = np.float32(conf_dict['spacing'].real)
    
#     cutoff_u = np.int(conf_dict['cutoff_u'].real)
#     cutoff_v = np.int(conf_dict['cutoff_v'].real)
#     u = np.array([np.float32(conf_dict['ux'].real), 
#                   np.float32(conf_dict['uy'].real), 
#                   np.float32(conf_dict['uz'].real)])
#     v = np.array([np.float32(conf_dict['vx'].real), 
#                   np.float32(conf_dict['vy'].real), 
#                   np.float32(conf_dict['vz'].real)])
    
#     for i in numba.prange(len(geo)):    # explicit parallel loop
#         R2 = geo[i]       # "observer"
#         for j in numba.prange(len(geo)):
#             R1 = geo[j]   # emitter
#             aj = alpha[j]
#             st = selfterms[j]
#             ## --- vacuum dyad
#             if i==j:
#                 ## self term
#                 xx, yy, zz = st[0,0], st[1,1], st[2,2]
#                 xy, xz, yx = st[0,1], st[0,2], st[1,0]
#                 yz, zx, zy = st[1,2], st[2,0], st[2,1]
#             else:
#                 xx, yy, zz, xy, xz, yx, yz, zx, zy = G0_periodic_123(
#                                         R1, R2, wavelength, eps1, eps2, eps3, 
#                                         spacing, k_x, cutoff_u, cutoff_v, u, v
#                                         )
            
#             ## --- 1-2-3 surface dyad (non retarded NF approximation)
#             if eps1!=eps2 or eps2!=eps3:
#                 xxs,yys,zzs,xys,xzs,yxs,yzs,zxs,zys = Gs_periodic_123(
#                                         R1, R2, wavelength, eps1, eps2, eps3, 
#                                         spacing, k_x, cutoff_u, cutoff_v, u, v
#                                         )
#                 ## combined dyad
#                 xx, yy, zz, xy, xz, yx, yz, zx, zy = xx+xxs, yy+yys, zz+zzs, \
#                                                       xy+xys, xz+xzs, yx+yxs, \
#                                                       yz+yzs, zx+zxs, zy+zys
            
#             ## return invertible matrix:  delta_ij*1 - G[i,j] * alpha[j]
#             M[3*i+0, 3*j+0] = -1*(xx*aj[0,0] + xy*aj[1,0] + xz*aj[2,0])
#             M[3*i+1, 3*j+1] = -1*(yx*aj[0,1] + yy*aj[1,1] + yz*aj[2,1])
#             M[3*i+2, 3*j+2] = -1*(zx*aj[0,2] + zy*aj[1,2] + zz*aj[2,2])
#             M[3*i+0, 3*j+1] = -1*(xx*aj[0,1] + xy*aj[1,1] + xz*aj[2,1])
#             M[3*i+0, 3*j+2] = -1*(xx*aj[0,2] + xy*aj[1,2] + xz*aj[2,2])
#             M[3*i+1, 3*j+0] = -1*(yx*aj[0,0] + yy*aj[1,0] + yz*aj[2,0])
#             M[3*i+1, 3*j+2] = -1*(yx*aj[0,2] + yy*aj[1,2] + yz*aj[2,2])
#             M[3*i+2, 3*j+0] = -1*(zx*aj[0,0] + zy*aj[1,0] + zz*aj[2,0])
#             M[3*i+2, 3*j+1] = -1*(zx*aj[0,1] + zy*aj[1,1] + zz*aj[2,1])
#             if i==j:
#                 M[3*i+0, 3*j+0] += 1
#                 M[3*i+1, 3*j+1] += 1
#                 M[3*i+2, 3*j+2] += 1


# @numba.njit(parallel=True, cache=True)
# def t_sbs_HE_123_quasistatic_periodic(geo, wavelength, selfterms, alpha, conf_dict, M):
#     k_x = np.float32(conf_dict['k_x'].real)
    
#     eps1 = conf_dict['eps1']
#     eps2 = conf_dict['eps2']
#     eps3 = conf_dict['eps3']
#     spacing = np.float32(conf_dict['spacing'].real)
    
#     cutoff_u = np.int(conf_dict['cutoff_u'].real)
#     cutoff_v = np.int(conf_dict['cutoff_v'].real)
#     u = np.array([np.float32(conf_dict['ux'].real), 
#                   np.float32(conf_dict['uy'].real), 
#                   np.float32(conf_dict['uz'].real)])
#     v = np.array([np.float32(conf_dict['vx'].real), 
#                   np.float32(conf_dict['vy'].real), 
#                   np.float32(conf_dict['vz'].real)])
    
#     for i in numba.prange(len(geo)):    # explicit parallel loop
#         R2 = geo[i]        # "observer"
#         st = selfterms[i]
#         for j in numba.prange(len(geo)):
#             R1 = geo[j]    # emitter
#             aj = alpha[j]
#             ## --- vacuum dyad
#             if i==j:
#                 ## self term
#                 xx, yy, zz = st[0,0], st[1,1], st[2,2]
#                 xy, xz, yx = st[0,1], st[0,2], st[1,0]
#                 yz, zx, zy = st[1,2], st[2,0], st[2,1]
#             else:
#                 xx, yy, zz, xy, xz, yx, yz, zx, zy = G_HE_periodic(
#                                         R1, R2, wavelength, eps1, eps2, eps3, 
#                                         spacing, k_x, cutoff_u, cutoff_v, u, v
#                                         )
            
#             ## return: G[i,j] * alpha[j]
#             ## --- magnetic-electric part
#             M[3*i+0, 3*j+0] = -1*(xx*aj[0,0] + xy*aj[1,0] + xz*aj[2,0])
#             M[3*i+1, 3*j+1] = -1*(yx*aj[0,1] + yy*aj[1,1] + yz*aj[2,1])
#             M[3*i+2, 3*j+2] = -1*(zx*aj[0,2] + zy*aj[1,2] + zz*aj[2,2])
#             M[3*i+0, 3*j+1] = -1*(xx*aj[0,1] + xy*aj[1,1] + xz*aj[2,1])
#             M[3*i+0, 3*j+2] = -1*(xx*aj[0,2] + xy*aj[1,2] + xz*aj[2,2])
#             M[3*i+1, 3*j+0] = -1*(yx*aj[0,0] + yy*aj[1,0] + yz*aj[2,0])
#             M[3*i+1, 3*j+2] = -1*(yx*aj[0,2] + yy*aj[1,2] + yz*aj[2,2])
#             M[3*i+2, 3*j+0] = -1*(zx*aj[0,0] + zy*aj[1,0] + zz*aj[2,0])
#             M[3*i+2, 3*j+1] = -1*(zx*aj[0,1] + zy*aj[1,1] + zz*aj[2,1])


### --- dyad class 3D non-retarded, 1D or 2D periodic
class DyadsQuasistaticPeriodic123(propagators.DyadsBaseClass):
    __name__ = "Quasistatic 3D '1-2-3' Green's tensors with periodicity"
    
    def __init__(self, n1=None, n2=None, n3=None, spacing=5000, 
                 u=np.array([0,0,0]), v=np.array([0,0,0]), 
                 cutoff_u=10, cutoff_v=10, 
                 radiative_correction=False,
                 n_cpu=-1):
        """ 
        u or v: period vectors
        set np.array([0,0,0]) for no periodicity in this dimension
        
        incident wavevector must be in XZ plane
        
        """
        super().__init__()
        
        self.u = u.astype(self.dtypef)
        self.v = v.astype(self.dtypef)
        
        self.cutoff_u = cutoff_u
        self.cutoff_v = cutoff_v
        if np.linalg.norm(u)==0: 
            self.cutoff_u = 0
        if np.linalg.norm(v)==0: 
            self.cutoff_v = 0
            
        ## Dyads
        self.G_EE = Gtot_periodic_123
        self.G_HE = G_HE_periodic
        self.G_EE_ff = Gtot_periodic_123    # using full propagator here for the moment
        
        ## evaluate propagator routine
        self.eval_G = greens_tensor_evaluation_periodic
        
        ## coupling matrix constructor routines
        self.tsbs_EE = t_sbs_EE_123_quasistatic_periodic
        self.tsbs_HE = t_sbs_HE_123_quasistatic_periodic
        
        ## environment definition
        ## set ref. index values or material class of environment layers
        from pyGDM2 import materials
        if isinstance(n1, (int, float, complex)) and not isinstance(n1, bool):
            self.n1_material = materials.dummy(n1)
        else:
            self.n1_material = n1
        
        n2 = n2 or n1     # if None, use `n1`
        if isinstance(n2, (int, float, complex)) and not isinstance(n2, bool):
            from pyGDM2 import materials
            self.n2_material = materials.dummy(n2)
        else:
            self.n2_material = n2
            
        n3 = n3 or n2     # if None, use `n2`
        if isinstance(n3, (int, float, complex)) and not isinstance(n3, bool):
            from pyGDM2 import materials
            self.n3_material = materials.dummy(n3)
        else:
            self.n3_material = n3
        
        self.spacing = spacing
        self.radiative_correction = radiative_correction
    
        self.n_cpu = n_cpu
        
    
    def __repr__(self, verbose=False):
        """description about simulation environment defined by set of dyads
        """
        out_str =  ' ------ environment -------'
        out_str += '\n ' + self.__name__
        out_str += '\n\n '
        out_str += 'peridicity: U={}, V={}\n '.format(self.u, self.v)
        out_str += '\n' + '   n3 = {}  <-- top'.format(self.n3_material.__name__)
        out_str += '\n' + '   n2 = {}  <-- structure zone (height "spacing" = {}nm)'.format(
                        self.n2_material.__name__, self.spacing)
        out_str += '\n' + '   n1 = {}  <-- substrate'.format(self.n1_material.__name__)
        return out_str

    
    def getConfigDictG(self, wavelength, struct, efield):
        if 'theta_inc' in efield.kwargs_permutations[0].keys():
            inc_angle = np.pi * efield.kwargs_permutations[0]['theta_inc'] / 180.0
        elif 'inc_angle' in efield.kwargs_permutations[0].keys():
            inc_angle = np.pi * efield.kwargs_permutations[0]['inc_angle'] / 180.0
        elif 'kSign' in efield.kwargs_permutations[0].keys():
            if efield.kwargs_permutations[0]['kSign'] == 1:
                inc_angle = 0.0    # bottom incidence
            else:
                inc_angle = np.pi  # top incidence
        else:
                inc_angle = np.pi  # top incidence
        
        
        ## all data need to be same dtype, must be cast to correct type inside numba functions
        conf_dict = numba.typed.Dict.empty(key_type=numba.types.unicode_type,
                                           value_type=numba.types.complex64)
        
        conf_dict['eps1'] = np.complex64(self.n1_material.epsilon(wavelength))
        conf_dict['eps2'] = np.complex64(self.n2_material.epsilon(wavelength))
        conf_dict['eps3'] = np.complex64(self.n3_material.epsilon(wavelength))
        conf_dict['spacing'] = np.complex64(self.spacing)
        conf_dict['ux'] = np.complex64(self.u[0])
        conf_dict['uy'] = np.complex64(self.u[1])
        conf_dict['uz'] = np.complex64(self.u[2])
        conf_dict['vx'] = np.complex64(self.v[0])
        conf_dict['vy'] = np.complex64(self.v[1])
        conf_dict['vz'] = np.complex64(self.v[2])
        conf_dict['cutoff_u'] = np.complex64(self.cutoff_u)
        conf_dict['cutoff_v'] = np.complex64(self.cutoff_v)
        
        conf_dict['n_cpu'] = np.complex64(self.n_cpu)
        
        n2 = (self.n2_material.epsilon(wavelength)**0.5)
        k_x = np.sin(inc_angle) * 2*np.pi * np.sqrt(n2**2) / wavelength
        conf_dict['k_x'] = np.complex64(k_x)
        
        ## return a numba typed dictionary of "complex64" type,
        ## can be used to pass configuration to the green's functions
        return conf_dict
    
    
    def exceptionHandling(self, struct, efield):
        """Exception handling / consistency and compatibility check
        
        check if structure and incident field generator are compatible

        Parameters
        ----------
        struct : :class:`.structures.struct`
            instance of structure class
        efield : :class:`.fields.efield`
            instance of incident field class

        Returns
        -------
        bool : True if struct and field are compatible, False if they don't fit the tensors

        """
        if efield.field_generator.__name__ not in ['plane_wave', 'planewave', 'evanescent_planewave']:
            raise Exception("periodic structures only work with plane wave illumination. " +
                            "Please use 'plane_wave', 'planewave' or 'evanescent_planewave' field generator!")
        
        all_inc_angles = []
        for k in efield.kwargs_permutations:
            if 'theta_inc' in k.keys():
                all_inc_angles.append(k['theta_inc'])
            if 'inc_plane' in k.keys():
                if k['inc_plane'] == 'yz':
                    raise Exception("'yz' plane of incidence is not supported by periodic structures. please use 'xz' incidence plane.")
        if len(np.unique(all_inc_angles)) > 1:
            raise Exception("For technical reasons periodic simulations can treat only a single incident angle. " + 
                            "Please run multiple separate simulations.")
        
        return True
    
    
    def getEnvironmentIndices(self, wavelength, geo):
        """get environment permittivity for `wavelength` at each meshpoint"""
        self.n1 = self.n1_material.epsilon(wavelength)**0.5
        self.n2 = self.n2_material.epsilon(wavelength)**0.5
        self.n3 = self.n3_material.epsilon(wavelength)**0.5
        
        ## environment epsilon at every meshpoint
        eps_env = np.zeros(len(geo), dtype=self.dtypec)
        eps_env[geo.T[2].min() > self.spacing] = self.n3_material.epsilon(wavelength)
        eps_env[0 <= geo.T[2].min() <= self.spacing] = self.n2_material.epsilon(wavelength)
        eps_env[geo.T[2].min() < 0] = self.n1_material.epsilon(wavelength)
        
        return eps_env
        
        
    def getSelfTermEE(self, wavelength, struct):
        eps_env = self.getEnvironmentIndices(wavelength, struct.geometry)
        struct.setWavelength(wavelength)
        
        k0 = 2.0*np.pi / float(wavelength)
        
        if struct.normalization == 0:
            cnorm = np.zeros(len(eps_env))
        else:
            norm_nonrad = -4.0 * np.pi * struct.normalization / (3.0 * struct.step**3 * eps_env)
            
            if self.radiative_correction:
                norm_rad = 1j * 2.0 * struct.normalization * (k0**3)/3.0 * np.ones(len(norm_nonrad))
                cnorm = norm_nonrad + norm_rad
            else:
                cnorm = norm_nonrad
        
        self_term_tensors_EE = np.zeros([len(norm_nonrad), 3, 3], dtype=self.dtypec)
        self_term_tensors_EE[:,0,0] = cnorm
        self_term_tensors_EE[:,1,1] = cnorm
        self_term_tensors_EE[:,2,2] = cnorm
        
        return self_term_tensors_EE
        
    
    def getSelfTermHE(self, wavelength, struct):
        eps_env = self.getEnvironmentIndices(wavelength, struct.geometry)
        struct.setWavelength(wavelength)
        
        self_term_tensors_HE = np.zeros([len(eps_env), 3, 3], dtype=self.dtypec)
        
        return self_term_tensors_HE
        
        
    def getPolarizabilityTensor(self, wavelength, struct):
        eps_env = self.getEnvironmentIndices(wavelength, struct.geometry)
        struct.setWavelength(wavelength)
        normalization = struct.normalization
        eps = struct.epsilon_tensor 
        
        vcell_norm = struct.step**3 / float(normalization)
        
        eps_env_tensor = np.zeros(eps.shape, dtype=self.dtypec)
        eps_env_tensor[:,0,0] = eps_env
        eps_env_tensor[:,1,1] = eps_env
        eps_env_tensor[:,2,2] = eps_env
        
        ## --- isotropic polarizability
        alphatensor = np.asfortranarray((eps - eps_env_tensor) * 
                                      vcell_norm / (4.0 * np.pi), dtype=self.dtypec)
        
        return alphatensor



#%% test
if __name__ == '__main__':
    from threadpoolctl import threadpool_limits
    from pyGDM2 import structures
    from pyGDM2 import materials
    from pyGDM2 import fields

    from pyGDM2 import core as core_old
    from pyGDM2 import core_py as core
    from pyGDM2 import visu
    from pyGDM2 import tools
    from pyGDM2 import linear

    from pyGDM2 import propagators_periodic_slow
    # from periodic_propagators import G_123_quasistatic_periodic


    from pyGDM2 import linear
    
    
    threadpool_limits(limits=4, user_api='blas')




    ## --------------- Setup structure
    mesh = 'cube'
    step = 25.0

    ## block 1: gold
    geom1 = structures.rect_wire(step, L=10,H=3,W=3, mesh=mesh)
    geom1 = structures.rect_wire(step, L=20,H=4,W=10, mesh=mesh)
    mat1 = len(geom1)*[materials.gold()]
    # mat1 = len(geom1)*[materials.silver()]
    mat1 = len(geom1)*[materials.dummy(3.5)]

    geom1.T[2] += 10

    geometry = geom1
    material = mat1 


    geometry = structures.center_struct(geometry)

    struct = structures.struct(step, geometry, material)

    ## incident field
    # field_generator = fields.planewave
    # kwargs = dict(theta=0)

    field_generator = fields.evanescent_planewave
    field_generator = fields.plane_wave
    kwargs = dict(E_s=0.3, E_p=1, inc_angle=25)


    wavelengths = [600]
    efield = fields.efield(field_generator, wavelengths=wavelengths, kwargs=kwargs)

    print("N_dp:", len(geometry))


    ## simulation initialization
    ## --- isolated structure
    n1 = 1.0
    n2 = 1.0
    dyads_single = propagators.DyadsQuasistatic123(n1=n1, n2=n2, spacing=1E9)
    sim1 = core.simulation(struct, efield, dyads_single)
    sim1.scatter(method='cupy')

    ## --- dyads: periodic
    cutoff=30
    I_nf_2 = []
    I_nf_3 = []
    cutofflist = np.arange(1,20,2)
    if 1:
    # for cutoff in cutofflist:
        print(cutoff)
        
        P = 800
        
        ## 1D periodic
        dyads_periodic_3 = propagators_periodic_slow.DyadsQuasistaticPeriodic123(u=np.array([80, 10, 0]), cutoff_u=cutoff, n1=n1, n2=n2, spacing=1E9)
        dyads_periodic_3 = propagators_periodic_slow.DyadsQuasistaticPeriodic123(
            u=np.array([P, 0, 0]), cutoff_u=cutoff, 
            v=np.array([0, P, 0]), cutoff_v=cutoff, 
            n1=n1, spacing=1E9)
        sim2 = core.simulation(struct, efield, dyads=dyads_periodic_3)
    
        ## new implementation 
        dyads_periodic_3_f = DyadsQuasistaticPeriodic123(u=np.array([80, 10, 0]), cutoff_u=cutoff, n1=n1, n2=n2, spacing=1E9)
        dyads_periodic_3_f = DyadsQuasistaticPeriodic123(
            u=np.array([P, 0, 0]), cutoff_u=cutoff, 
            v=np.array([0, P, 0]), cutoff_v=cutoff,
            n1=n1, spacing=1E9, n_cpu=-1)
        sim3 = core.simulation(struct, efield, dyads=dyads_periodic_3_f)
        # dyads_periodic_81 = propagators.G_123_quasistatic_periodic(u=np.array([1000, 0, 0]), cutoff_u=40)
        # sim3 = core.simulation(struct, efield, dyads=dyads_periodic_81)
    
        
    
        # visu.structure(sim1)
        ##%%
        # sim2.scatter(method='cupy')
        sim3.scatter(method='cupy')
        
        ##%%
        import matplotlib.pyplot as plt
        plt.figure(figsize=(15,5))
        plt.subplot(131)
        visu.vectorfield_by_fieldindex(sim1, 0, show=0)
        plt.subplot(132)
        visu.vectorfield_by_fieldindex(sim2, 0, show=0)
        plt.subplot(133)
        visu.vectorfield_by_fieldindex(sim3, 0, show=0)
        plt.show()
        
        r_probe = tools.generate_NF_map(-100, 100, 21, 0, 0, 1, 200)
        print ('calc NF, sim2...')
        Es, Et, Bs, Bt = linear.nearfield(sim2, 0, r_probe)
        I2=np.sum(np.abs(Et.T[3:]), axis=0)
        print ('calc NF, sim3...')
        Es, Et, Bs, Bt = linear.nearfield(sim3, 0, r_probe)
        I3=np.sum(np.abs(Et.T[3:]), axis=0)
        
        plt.plot(I2)
        plt.plot(I3)
        
        I_nf_2.append(I2)
        I_nf_3.append(I3)

#%%
    I_nf_2 = np.array(I_nf_2)
    I_nf_3 = np.array(I_nf_3)
    
    plt.subplot(131)
    plt.imshow(I_nf_2)
    plt.colorbar()
    plt.subplot(132)
    plt.imshow(I_nf_3)
    plt.colorbar()
    plt.subplot(133)
    plt.imshow((I_nf_3-I_nf_2)/I_nf_3, cmap='bwr')
    plt.colorbar()
    plt.show()
    
    Es, Et, Bs, Bt = linear.nearfield(sim1, 0, r_probe)
    I1=np.sum(np.abs(Et.T[3:]), axis=0)
    plt.plot(I1)
    plt.show()
    #%% test convergence
    


    x = 31.
    y = 51.
    z = 15.
    
    wavelength = 550.
    eps1= eps2 = eps3 = complex(1)
    spacing = 10000.
    k_x = 0.0
    
    u=np.array([0., 400, 0])
    v=np.array([400, 0., 0])
    # v=np.array([0, 0, 0])
    
    
    
    Gtest = []
    Gtest2 = []
    for i in range(1, 101):
        
        cutoff_u = cutoff_v = i
        if np.linalg.norm(v) == 0:
            cutoff_v=0
        
        
        R1 = np.array([0,0,0])
        R2 = np.array([x,y,z])
        xx, yy, zz, xy, xz, yx, yz, zx, zy = G0_periodic_123(
                        R1, R2, wavelength, eps1, eps2, eps3, 
                        spacing, k_x, cutoff_u, cutoff_v, u, v
                        )
        _G = np.array([[xx,xy,xz], [yx,yy,yz], [zx,zy,zz]], dtype=np.complex64)
        
        xx, yy, zz, xy, xz, yx, yz, zx, zy = G0_periodic_123_direct_sum(
                        R1, R2, wavelength, eps1, eps2, eps3, 
                        spacing, k_x, cutoff_u, cutoff_v, u, v
                        )
        _G2 = np.array([[xx,xy,xz], [yx,yy,yz], [zx,zy,zz]], dtype=np.complex64)

        Gtest.append(_G)
        Gtest2.append(_G2)
        
        
    
    Gtest = np.array(Gtest)
    Gtest2 = np.array(Gtest2)
    # Gtest = np.abs(Gtest) / np.linalg.norm(np.abs(Gtest[10:]), axis=(1,2)).max()
    # Gtest2 = np.abs(Gtest2) / np.linalg.norm(np.abs(Gtest2[10:]), axis=(1,2)).max()
    
    
    idx = [0,1]
    # mv_avg = moving_average(Gtest[:,idx[0],idx[1]], n=n_mvavg)
    # plt.plot(mv_avg[:])
    plt.plot(Gtest[1:,idx[0],idx[1]].real, color='C0', label='moving avg.')
    plt.plot(Gtest2[1:,idx[0],idx[1]].real, color='C1', label='simple sum')
    plt.axhline(np.mean(Gtest[-10:, idx[0],idx[1]].real), dashes=[3,3], color='C0', lw=1, label='last 10 average')
    plt.axhline(np.mean(Gtest2[-10:, idx[0],idx[1]].real), dashes=[3,3], color='C1', lw=1, label='last 10 average')
    # plt.ylim(-0.5,0.5)
    plt.legend()
    plt.xlabel('cutoff period')
    plt.ylabel('Re(Gxy)')
    plt.show()
    #%% test 2
    ### --- 3D periodic Dyads - until now homogeneous environment only, normal incidence only
    @numba.njit(parallel=True, cache=True)
    def G0_periodic_123cumsum(R1, R2, wavelength, 
                        eps1=1.0+0j, eps2=1.0+0j, eps3=1.0+0j, spacing=5000, 
                        k_x=0.0, N_u=0, N_v=0, 
                        u=np.array([0., 0., 0.]), v=np.array([0., 0., 0.])):
        xx = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
        yy = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
        zz = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
        xy = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
        xz = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
        yx = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
        yz = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
        zx = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
        zy = np.zeros((N_u*2 + 1, N_v*2 + 1), dtype=np.complex128)
        
        for _u_i in numba.prange(1, 2*(N_u+1)):
            u_i = (_u_i%2-0.5)*2 * _u_i//2
            for _v_i in numba.prange(1, 2*(N_v+1)):
                v_i = (_v_i%2-0.5)*2 * _v_i//2 
                _R1 = R1 + u*u_i + v*v_i     # position of emitter-copy from periodic cell [u_i, v_i]
                _xx, _yy, _zz, _xy, _xz, _yx, _yz, _zx, _zy = propagators.G0_EE_123(
                        _R1, R2, wavelength, eps1, eps2, eps3, spacing)
                xx[_u_i-1, _v_i-1] = _xx
                yy[_u_i-1, _v_i-1] = _yy
                zz[_u_i-1, _v_i-1] = _zz
                xy[_u_i-1, _v_i-1] = _xy
                xz[_u_i-1, _v_i-1] = _xz
                yx[_u_i-1, _v_i-1] = _yx
                yz[_u_i-1, _v_i-1] = _yz
                zx[_u_i-1, _v_i-1] = _zx
                zy[_u_i-1, _v_i-1] = _zy
        
        ## calc. mean of moving average of tail of series to dampen oscillations
        # xx = np.mean(moving_average(xx.flatten(), n=N_u*N_v)[2*N_u*N_v:-1])
        # yy = np.mean(moving_average(yy.flatten(), n=N_u*N_v)[2*N_u*N_v:-1])
        # zz = np.mean(moving_average(zz.flatten(), n=N_u*N_v)[2*N_u*N_v:-1])
        # xy = np.mean(moving_average(xy.flatten(), n=N_u*N_v)[2*N_u*N_v:-1])
        # xz = np.mean(moving_average(xz.flatten(), n=N_u*N_v)[2*N_u*N_v:-1])
        # yx = np.mean(moving_average(yx.flatten(), n=N_u*N_v)[2*N_u*N_v:-1])
        # yz = np.mean(moving_average(yz.flatten(), n=N_u*N_v)[2*N_u*N_v:-1])
        # zx = np.mean(moving_average(zx.flatten(), n=N_u*N_v)[2*N_u*N_v:-1])
        # zy = np.mean(moving_average(zy.flatten(), n=N_u*N_v)[2*N_u*N_v:-1])
        
        return xx, yy, zz, xy, xz, yx, yz, zx, zy
        # return np.cumsum(xx), np.cumsum(yy), np.cumsum(zz), \
        #         np.cumsum(xy), np.cumsum(xz), np.cumsum(yx), \
        #         np.cumsum(yz), np.cumsum(zx), np.cumsum(zy)
               
           
    u=np.array([0., 600, 0])
    v=np.array([500, 0., 0])
    n_mvavg = 10
    i0_sample, i1_sample = 10,25
        
    cutoff_u = cutoff_v = 50
    t0=time.time()
    xx, yy, zz, xy, xz, yx, yz, zx, zy = G0_periodic_123cumsum(
                    R1, R2, wavelength, eps1, eps2, eps3, 
                    spacing, k_x, cutoff_u, cutoff_v, u, v
                    )
    print(time.time()-t0)
    
    # test = zx
    # plt.plot(np.cumsum(test.flatten()))
    
    # csum = moving_average(np.cumsum(test.flatten()), n=cutoff_u*cutoff_v)
    # plt.plot(csum)
    # plt.plot(moving_average(test.flatten(), n=cutoff_u*cutoff_v)[10:])
    # plt.axhline(0, color='k')
    # plt.axhline(np.mean(csum[2*cutoff_u*cutoff_v:-1]), color='C0', dashes=[3,3])
    # plt.axvline(2*cutoff_u*cutoff_v, color='k')
    # plt.ylim(-1E-8, 1E-8)
    


# encoding: utf-8
#
#Copyright (C) 2017-2024, P. R. Wiecha
#
#This program is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
nonlinear optical effects

"""
from __future__ import print_function
from __future__ import absolute_import

import numpy as np
import copy



# ## (on linux) Set Stacksize for enabling passing large arrays to the fortran subroutines
# import platform
# if platform.system() == 'Linux':
#     import resource
#     resource.setrlimit(resource.RLIMIT_STACK, (resource.RLIM_INFINITY, resource.RLIM_INFINITY))


#==============================================================================
# GLOBAL PARAMETERS
#==============================================================================




#==============================================================================
# EXCEPTIONS
#==============================================================================








#==============================================================================
# incoherent nonlinear effects
#==============================================================================
def tpl_ldos(sim, field_index, nonlin_order=2.0, beta=1.0E5, verbose=False):
    """calculate the TPL for nano-object under given illumination condition
    
    Calculate the two-photon photolouminescence (TPL). Higher order nonlinear 
    photoluminescence can also be calculated by changing the parameter `nonlin_order`.
    
     - Can be used to simulate TPL signals using `nonlin_order`=2 (default)
      
     - Using `nonlin_order`=1 together with an unphysically tight 
       focused beam --> calculate the LDOS. 
       WARNING: The focus waist must not be smaller than some times the stepsize!
    
     - Might be used for Raman intensities using `nonlin_order`=1.
    
    
    Parameters
    ----------
    sim : :class:`.core.simulation`
        simulation description
    
    field_index : int
        index of evaluated self-consistent field to use for calculation. Can be
        obtained for specific parameter-set using :func:`.tools.get_closest_field_index`
    
    nonlin_order : float, default: 2.0
        order of nonlinear response. (I_npl ~ | |E|^2 | ^ nonlin_order).
    
    beta : float, default: 1E5
        incident laser power scaling. Use 1E5 for correct LDOS values.
    
    verbose : bool, default: False
        Enable some info printing
    
    
    Returns
    -------
    I_tpl : float
        TPL intensity in farfield from nano-object under given illumination
    
    
    Notes
    -----
    For details on TPL/LDOS calculation via focused beam rasterscan 
    simulations, see:
    Viarbitskaya, S. et al. **Tailoring and imaging the plasmonic local 
    density of states in crystalline nanoprisms**, Nat. Mater. 12, 426–432 (2013)

    """
    if sim.E is None: 
        raise ValueError("Error: Scattering field inside the structure not yet evaluated. Run `core.scatter` simulation first.")
    
    if verbose:
        if nonlin_order == 1:
            print("nonlinear order = 1 --> Using this is adapted for LDOS.")
        if nonlin_order == 2:
            print("nonlinear order = 2 --> Using this is adapted for TPL signal.")
        if nonlin_order > 2:
            print("nonlinear order > 2 --> Higher order incoherent nonlinear luminescence?")
        
    field_params = sim.E[field_index][0]
    wavelength   = field_params['wavelength']
    k0 = 2*np.pi / wavelength
    
    ## --- environment
    _ = sim.dyads.getEnvironmentIndices(wavelength, sim.struct.geometry)
    k1 = k0 * np.real(sim.dyads.n2)
    
    ## --- constant factors
    pre_factor = (16.*(k1**4))/3.
    
    ## --- meshpoint electric fields
    E = sim.E[field_index][1]
    ex, ey, ez = E.T
    
    ## --- total TPL intensity
    I_e = np.power((np.abs(ex)**2 + np.abs(ey)**2 + np.abs(ez)**2), nonlin_order)
    I_tpl = beta * pre_factor * np.sum(I_e)
    
    return I_tpl




#==============================================================================
# coherent nonlinear effects
#==============================================================================


## --- NONE SO FAR.







# encoding: utf-8
from pyGDM2 import structures
from pyGDM2 import materials
from pyGDM2 import propagators
from pyGDM2 import fields
from pyGDM2 import core
from pyGDM2 import visu

## --- alternatively use new API:
# from pyGDM2 import fields_py as fields
# from pyGDM2 import core_py as core

## --- simulation initialization ---
## structure: sphere of 160nm radius,
## constant dielectric function,
step = 20
geometry = structures.sphere(step, R=8.2, mesh='cube')
material = materials.dummy(2.0)
struct = structures.struct(step, geometry, material)

## incident field: plane wave, 400nm, lin. pol. along X
field_generator = fields.plane_wave
wavelengths = [400]
kwargs = dict(inc_angle=0, inc_plane='xz', theta=0)
efield = fields.efield(field_generator, 
               wavelengths=wavelengths, kwargs=kwargs)

## environment: vacuum
n1 = 1.0
dyads = propagators.DyadsQuasistatic123(n1)

## create simulation object
sim = core.simulation(struct, efield, dyads)


## --- run the simulation ---
core.scatter(sim, method='cupy')


## --- plot the near-field inside the sphere ---
## using first (of one) field-config (=index 0)
## slice through sphere center
visu.vectorfield_color_by_fieldindex(sim, 0, projection='XY', slice_level=160)
visu.vectorfield_color_by_fieldindex(sim, 0, projection='XZ', slice_level=0)
visu.vectorfield_color_by_fieldindex(sim, 0, projection='YZ', slice_level=0)





# encoding: utf-8
#
#Copyright (C) 2017-2024, P. R. Wiecha
#
#This program is free software: you can redistribute it and/or modify
#it under the terms of the GNU General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.
#
#This program is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU General Public License for more details.
#
#You should have received a copy of the GNU General Public License
#along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
tools around multipole decomposition and effective polarizabilities

"""
from __future__ import print_function
from __future__ import absolute_import

import warnings
# import math

import numpy as np
import copy
import numba

import time

from pyGDM2.multipole.main import multipole_decomposition_exact
from pyGDM2.multipole.polarizabilities import eval_generalized_polarizability_m
from pyGDM2.multipole.polarizabilities import eval_generalized_polarizability_p
from pyGDM2.multipole.polarizabilities import eval_generalized_polarizability_qe
from pyGDM2.multipole.polarizabilities import eval_generalized_polarizability_qm
from pyGDM2.multipole.polarizabilities import _test_availability_generalized_polarizability
from pyGDM2.multipole import propagators as mp_propagators

#==============================================================================
# GLOBAL PARAMETERS
#==============================================================================
DTYPE_C = np.complex64




#==============================================================================
# EXCEPTIONS
#==============================================================================



# =============================================================================
# CONSISTENCY TESTS / HELPER
# =============================================================================
def _check_multipole_list(p, m, qe, qm, r_p, r_m, r_qe, r_qm, r0):
    
    # dipole moment definitions
    if len(p) == 0 and len(m) == 0 and len(qe) == 0 and len(qm) == 0:
        raise Exception("All multipoles are zero. Please define at least one multipole moment.")
    if len(p) == 0:
        p = np.zeros((0, 3), dtype=np.complex64)
        r_p = np.zeros((0, 3), dtype=np.float32)
    if len(m) == 0:
        m = np.zeros((0, 3), dtype=np.complex64)
        r_m = np.zeros((0, 3), dtype=np.float32)
    if len(qe) == 0:
        qe = np.zeros((0, 3), dtype=np.complex64)
        r_qe = np.zeros((0, 3), dtype=np.float32)
    if len(qm) == 0:
        qm = np.zeros((0, 3), dtype=np.complex64)
        r_qm = np.zeros((0, 3), dtype=np.float32)
        
    # positions 
    if r0 is None:
        if r_p is None and r_m is None and r_qe is None and r_qm is None:
            raise Exception("No positions for the multipoles are given.")
        if len(r_p) != len(p):
            raise Exception("Same number of dipole positions and dipole moments required! `p` and `r_p` must be lists of same length.")
        if len(r_m) != len(m):
            raise Exception("Same number of dipole positions and dipole moments required! `m` and `r_m` must be lists of same length.")
        if len(r_qe) != len(qe):
            raise Exception("Same number of quadrupole positions and quadrupole moments required! `qe` and `r_qe` must be lists of same length.")
        if len(r_qm) != len(qm):
            raise Exception("Same number of quadrupole positions and quadrupole moments required! `qm` and `r_qm` must be lists of same length.")
    else:
        # global r0 defined: override specific positions
        r0 = np.array(r0)
        if len(r0) != 3: 
            raise Exception("Multipole position must be carthesian coordinate (array of 3 float)")
        if len(p) > 0:
            r_p = np.array([r0 for _ in p], dtype= np.float32)
        if len(m) > 0:
            r_m = np.array([r0 for _ in m], dtype= np.float32)
        if len(qe) > 0:
            r_qe = np.array([r0 for _ in qe], dtype= np.float32)
        if len(qm) > 0:
            r_qm = np.array([r0 for _ in qm], dtype= np.float32)
    
    return p, m, qe, qm, r_p, r_m, r_qe, r_qm, r0


def _eval_G0(G0func, R1, R2, wl, eps_env):
    """evaluate a free-space Green's tensor `G0func`
    """
    if np.linalg.norm(R1-R2)==0:
        warnings.warn("Green's tensor singularity. Returning zero.")
        return np.zeros((3,3), dtype=np.complex64)
    
    xx, yy, zz, xy, xz, yx, yz, zx, zy = G0func(R1, R2, wl, eps_env)
    G0 = np.array([
        [xx, xy, xz],
        [yx, yy, yz],
        [zx, zy, zz],
        ]).astype(np.complex64)
    
    return G0


# =============================================================================
# LOW-LEVEL (working with lists of multipoles)
# =============================================================================
def _multipole_extinct(sim, field_index,
                       p=[], m=[], qe=[], qm=[],
                       r_p=None, r_m=None, r_qe=None, r_qm=None,
                       r0=None, eps_dd=0.001, 
                       normalization_E0=True,
                       ):
    """extinction cross-section for an explicit list of multipole moments
    
    
    Parameters
    ----------
    wavelength : float
        radiation wavelength
    
    eps_env : float
        envirnoment permittivity (real part)
        
    r0 : array. optional
        [x,y,z] position of mulipole decomposition development.
        Alternatively the positions of each mutlipole can be specified.
        The default is `None`.
        
    p, m, qe, qm : lists of ndarrays. optional
        lists of dipole and quadrupole moments to propagate to the farfield.
        The default is [].
        
    r_p, r_m, r_qe, r_qm : ndarrays or list of such. optional
        Has effect only if r0 is not specified. Carthesian coordinates of 
        each multipole individually.
    
    
    Returns
    -------
    total extinction cross sections for each multipole type (in nm^2)
        
    
    Notes
    -----
    For details about the extinction section of multipole moments, see:
    
    Evlyukhin, A. B. et al. *Multipole analysis of light scattering by 
    arbitrary-shaped nanoparticles on a plane surface.*, 
    JOSA B 30, 2589 (2013)
    """
# =============================================================================
#     exception handling
# =============================================================================
    p, m, qe, qm, r_p, r_m, r_qe, r_qm, r0 = _check_multipole_list(
        p=p, m=m, qe=qe, qm=qm, r_p=r_p, r_m=r_m, r_qe=r_qe, r_qm=r_qm, r0=r0)
    
    
# =============================================================================
#     preparation
# =============================================================================  
    from pyGDM2 import tools 
    from pyGDM2 import linear
    
    field_params = tools.get_field_indices(sim)[field_index]
    wavelength   = field_params['wavelength']
    sim.struct.setWavelength(wavelength)
    k0 = 2*np.pi / wavelength
    eps_env = sim.dyads.getEnvironmentIndices(wavelength, sim.struct.geometry[:1])[0]  # structure must be fully in one environment zone
    n_env = (eps_env**0.5).real
    
    ## incident field and its gradients at multipole position
    env_dict = sim.dyads.getConfigDictG(wavelength, sim.struct, sim.efield)
    E0 = sim.efield.field_generator(r_p, env_dict, **field_params)
    H0 = sim.efield.field_generator(r_m, env_dict, returnField='H', **field_params)
    
    gradE0 = linear.field_gradient(sim, field_index, r_qe, 
                                   delta=eps_dd, which_fields=['E0'])
    gradH0 = linear.field_gradient(sim, field_index, r_qm, 
                                   delta=eps_dd, which_fields=['H0'])
    gradE0cj = np.conj(np.array(gradE0)[...,3:])
    gradH0cj = np.conj(np.array(gradH0)[...,3:])
    
    ## normalization
    if normalization_E0:
        # intensity of incident field at each multipole
        E2in_p = np.sum(np.abs(sim.efield.field_generator(r_p, env_dict, **field_params))**2, axis=1).real
        E2in_m = np.sum(np.abs(sim.efield.field_generator(r_m, env_dict, **field_params))**2, axis=1).real
        E2in_qe = np.sum(np.abs(sim.efield.field_generator(r_qe, env_dict, **field_params))**2, axis=1).real
        E2in_qm = np.sum(np.abs(sim.efield.field_generator(r_qm, env_dict, **field_params))**2, axis=1).real
    else:
        E2in_p = np.ones(len(r_p))
        E2in_m = np.ones(len(r_m))
        E2in_qe = np.ones(len(r_qe))
        E2in_qm = np.ones(len(r_qm))
    
    E0 = E0 / E2in_p[:,None]
    H0 = H0 / E2in_m[:,None]
    gradE0cj = gradE0cj / E2in_qe[:,None]
    gradH0cj = gradH0cj / E2in_qm[:,None]
    
# =============================================================================
#     calc the ECS per multipole order
# =============================================================================
    prefactor = (4 * np.pi * k0 * 1/n_env).real
    
    ecs_p = prefactor * (np.sum(np.conjugate(E0)*p)).imag
    ecs_m = prefactor / n_env * (np.sum(np.conjugate(H0)*m)).imag
    ecs_Qe = prefactor / 12. * (np.sum([np.tensordot(gradE0cj[:,i,:] + gradE0cj[:,i,:].T, qe[i]) for i in range(len(qe))] )).imag
    ecs_Qm = prefactor / n_env / 6 * (np.sum([np.tensordot(gradH0cj[:,i,:].T, qm[i]) for i in range(len(qm))] )).imag
    
    return [ecs_p, ecs_m, ecs_Qe, ecs_Qm]


def _multipole_scs(sim, field_index,
                   p=[], m=[], qe=[], qm=[],
                   r_p=None, r_m=None, r_qe=None, r_qm=None,
                   r0=None, eps_dd=0.001, 
                   normalization_E0=True,
                   ):
    """scattering cross-section for an explicit list of multipole moments
    
    Radiated electro-magnetic field  (E-field) in the far-field 
    (by default on a sphere of radius `r` around the nanostructure).
    
    
    Parameters
    ----------
    wavelength : float
        radiation wavelength
    
    eps_env : float
        envirnoment permittivity (real part)
        
    r0 : array. optional
        [x,y,z] position of mulipole decomposition development.
        Alternatively the positions of each mutlipole can be specified.
        The default is `None`.
        
    p, m, qe, qm : lists of ndarrays. optional
        lists of dipole and quadrupole moments to propagate to the farfield.
        The default is [].
        
    r_p, r_m, r_qe, r_qm : ndarrays or list of such. optional
        Has effect only if r0 is not specified. Carthesian coordinates of 
        each multipole individually.
    
    Returns
    -------
    total scattering cross sections for each multipole type (in nm^2)
        
    
    Notes
    -----
    For details about the exact multipole formalism and scs calculation, see: 
        
    Alaee, R., Rockstuhl, C. & Fernandez-Corbaton, I. *An electromagnetic 
    multipole expansion beyond the long-wavelength approximation.*
    Optics Communications 407, 17–21 (2018)        
    """
# =============================================================================
#     exception handling
# =============================================================================
    p, m, qe, qm, r_p, r_m, r_qe, r_qm, r0 = _check_multipole_list(
        p=p, m=m, qe=qe, qm=qm, r_p=r_p, r_m=r_m, r_qe=r_qe, r_qm=r_qm, r0=r0)
    
# =============================================================================
#     preparation
# =============================================================================  
    from pyGDM2 import tools 
    
    field_params = tools.get_field_indices(sim)[field_index]
    wavelength   = field_params['wavelength']
    sim.struct.setWavelength(wavelength)
    k0 = 2*np.pi / wavelength
    eps_env = sim.dyads.getEnvironmentIndices(wavelength, sim.struct.geometry[:1])[0]  # structure must be fully in one environment zone
    n_env = (eps_env**0.5).real
    k = k0*n_env
    
    ## normalization
    if normalization_E0:
        ## incident field intensity at each multipole location
        env_dict = sim.dyads.getConfigDictG(wavelength, sim.struct, sim.efield)
        
        E2in_p = np.sum(np.abs(sim.efield.field_generator(r_p, env_dict, **field_params))**2, axis=1).real
        E2in_m = np.sum(np.abs(sim.efield.field_generator(r_m, env_dict, **field_params))**2, axis=1).real
        E2in_qe = np.sum(np.abs(sim.efield.field_generator(r_qe, env_dict, **field_params))**2, axis=1).real
        E2in_qm = np.sum(np.abs(sim.efield.field_generator(r_qm, env_dict, **field_params))**2, axis=1).real
    else:
        E2in_p = np.ones(len(r_p))
        E2in_m = np.ones(len(r_m))
        E2in_qe = np.ones(len(r_qe))
        E2in_qm = np.ones(len(r_qm))
    
# =============================================================================
#     calc the ECS per multipole order
# =============================================================================
    ## factor 100: cm --> m (cgs units)
    sc_factor_dp = 100/12*(k0**4).real
    sc_factor_Q = 100/1440*(k0**4).real
        
    scs_p = sc_factor_dp * np.sum(np.abs(p)**2 / E2in_p[:,None])
    scs_m = sc_factor_dp * np.sum(np.abs(m)**2 / E2in_m[:,None])
    scs_Qe = sc_factor_Q * np.sum(np.abs(k*qe)**2 / E2in_qe[:,None])
    scs_Qm = sc_factor_Q * np.sum(np.abs(k*qm)**2 / E2in_qm[:,None])
    
    return [scs_p, scs_m, scs_Qe, scs_Qm]


def _multipole_nearfield(wavelength, r_probe, eps_env=1, 
                         p=[], m=[], qe=[], qm=[],
                         r_p=None, r_m=None, r_qe=None, r_qm=None,
                         r0=None, which_fields=["Es","Bs"]):
    """calculate scattered field of multiple multipole (valid in the nearfield region)
    
    !! not performance optimized !!
    
    Parameters
    ----------
    wavelength : float
        radiation wavelength
        
    r_probe : tuple (x,y,z) or list of 3-lists/-tuples
        (list of) coordinate(s) to evaluate nearfield on. 
        Format: tuple (x,y,z) or list of 3 lists: [Xmap, Ymap, Zmap] 
        (the latter can be generated e.g. using :func:`.tools.generate_NF_map`)
    eps_env : float
        envirnoment permittivity (real part)
        
    r0 : array. optional
        [x,y,z] position of mulipole decomposition development.
        Alternatively the positions of each mutlipole can be specified.
        The default is `None`.
        
    p, m, qe, qm : lists of ndarrays. optional
        lists of dipole and quadrupole moments to propagate to the farfield.
        The default is [].
        
    r_p, r_m, r_qe, r_qm : ndarrays or list of such. optional
        Has effect only if r0 is not specified. Carthesian coordinates of 
        each multipole individually.
        
    which_fields : list of str, default: ["Es","Bs"]
        which fields to calculate and return. available options: 
            ["Es", "Bs"]
    
    Returns
    -------
    depending on kwarg `which_fields`, one or two lists of 6-tuples, complex : 
        - scattered Efield ("Es")
        - scattered Bfield ("Bs")
    
    the tuples are of shape (X,Y,Z, Ax,Ay,Az) with Ai the corresponding 
    complex field component
    
    Notes
    -----
    Details about the calculation of the field scattered by quadrupoles:
    Babicheva and Evlyukhin
    **Analytical model of resonant electromagnetic dipole-quadrupole coupling 
    in nanoparticle arrays**
    Phys. Rev. B 99, 195444 (2019)

    """
# =============================================================================
#     exception handling
# =============================================================================
    p, m, qe, qm, r_p, r_m, r_qe, r_qm, r0 = _check_multipole_list(
        p=p, m=m, qe=qe, qm=qm, r_p=r_p, r_m=r_m, r_qe=r_qe, r_qm=r_qm, r0=r0)

# =============================================================================
#     the main calculation
# =============================================================================
    return_field_list = []
    for returnfield in which_fields:
        
        if returnfield.lower() in ['e', 'es']:
            func_G_p = mp_propagators.G0_Ep
            func_G_m = mp_propagators.G0_Em
            func_G_qe = mp_propagators.G0_Eqe
            func_G_qm = mp_propagators.G0_Eqm
        elif returnfield.lower() in ['h', 'b', 'hs', 'bs']:
            func_G_p = mp_propagators.G0_Hp
            func_G_m = mp_propagators.G0_Hm
            func_G_qe = mp_propagators.G0_Hqe
            func_G_qm = mp_propagators.G0_Hqm
            
        ## evaluate the propagators
        Fp = np.zeros([len(r_probe), 3], dtype=np.complex64)
        Fm = np.zeros([len(r_probe), 3], dtype=np.complex64)
        FQe = np.zeros([len(r_probe), 3], dtype=np.complex64)
        FQm = np.zeros([len(r_probe), 3], dtype=np.complex64)
        for i, _r in enumerate(r_probe):
            # ED
            for j, r0 in enumerate(r_p):
                n_vec = (_r - r0) / np.linalg.norm((_r - r0)) # scat. direction
                Gp = _eval_G0(func_G_p, r0, _r, wavelength, eps_env)
                Fp[i] += np.dot(Gp, p[j]) # add scattered field
                
            # MD
            for j, r0 in enumerate(r_m):
                n_vec = (_r - r0) / np.linalg.norm((_r - r0)) # scat. direction                
                Gm = _eval_G0(func_G_m, r0, _r, wavelength, eps_env)
                Fm[i] += np.dot(Gm, m[j]) # add scattered field
                
            # QE
            for j, r0 in enumerate(r_qe):
                n_vec = (_r - r0) / np.linalg.norm((_r - r0)) # scat. direction                
                Gqe = _eval_G0(func_G_qe, r0, _r, wavelength, eps_env)
                qen = np.tensordot(qe, n_vec, axes=(-1,-1))
                FQe[i] += np.dot(Gqe, qen[j]) # add scattered field
            
            # QM    
            for j, r0 in enumerate(r_qm):
                Gqm = _eval_G0(func_G_qm, r0, _r, wavelength, eps_env)
                qmn = np.tensordot(qm, n_vec, axes=(-1,-1))
                FQm[i] += np.dot(Gqm, qmn[j]) # add scattered field
            
        
        # sum contributions of all multipoles and append calculation positions
        F = Fp + Fm + FQe + FQm
        F_iwth_pos = np.concatenate([r_probe, F], axis=1)
        return_field_list.append(F_iwth_pos)
    
    return return_field_list


def _multipole_farfield(wavelength, eps_env, r0=None, 
                        p=[], m=[], qe=[], qm=[],
                        r_p=None, r_m=None, r_qe=None, r_qm=None,
                        r_probe=None,
                        r=100000., 
                        tetamin=0, tetamax=np.pi/2., Nteta=10, 
                        phimin=0, phimax=2*np.pi, Nphi=36, 
                        return_value='map'):
    """far-field scattering of an explicit list of multipole moments
    
    Radiated electro-magnetic field  (E-field) in the far-field 
    (by default on a sphere of radius `r` around the nanostructure).
    
    
    Parameters
    ----------
    wavelength : float
        radiation wavelength
    
    eps_env : float
        envirnoment permittivity (real part)
        
    r0 : array. optional
        [x,y,z] position of mulipole decomposition development.
        Alternatively the positions of each mutlipole can be specified.
        The default is `None`.
        
    p, m, qe, qm : lists of ndarrays. optional
        lists of dipole and quadrupole moments to propagate to the farfield.
        The default is [].
        
    r_p, r_m, r_qe, r_qm : ndarrays or list of such. optional
        Has effect only if r0 is not specified. Carthesian coordinates of 
        each multipole individually.
    
    r_probe : tuple (x,y,z) or list of 3-lists/-tuples. optional
        defaults to *None*, which means it is not used and a solid angle defined by
        a spherical coordinate range is used instead. If `r_probe` is given, this
        overrides `r`, `tetamin`, `tetamax`, `Nteta`, `Nphi`.
        (list of) coordinate(s) to evaluate farfield on. 
        Format: tuple (x,y,z) or list of 3 lists: [Xmap, Ymap, Zmap] 
        The default is None. (Not used)
        
    r : float, default: 100000.
        radius of integration sphere (distance to coordinate origin in nm)
        
    tetamin, tetamax : float, float; defaults: 0, np.pi/2
        minimum and maximum polar angle in radians 
        (in linear steps from `tetamin` to `tetamax`)
        
    phimin, phimax : float, float; defaults: 0, 2*np.pi
        minimum and maximum azimuth angle in radians, excluding last position
        (in linear steps from `phimin` to `phimax`)
        
    Nteta, Nphi : int, int; defaults: 10, 36
        number of polar and azimuthal angles on sphere to calculate,
    
    return_value : str, default: 'map'
        Values to be returned. Either 'map' (default) or 'integrated'.
          - "map" : (default) return spatially resolved farfield intensity at each spherical coordinate (5 lists)
          - "efield" : return spatially resolved E-fields at each spherical coordinate (5 lists)
          - "int_Es" : return the integrated scattered field (as float)
    
    
    Returns
    -------
    using `r_probe` for position definition:
        3 lists of 6-tuples (x,y,z, Ex,Ey,Ez), complex : 
            - scattered Efield or E-intensity
        
    if solid angle is defined via spherical coordinate range:
        - return_value == "map" : 5 arrays of shape (Nteta, Nphi) : 
            - [tetalist : teta angles - if solid angle range input]
            - [philist : phi angles - if solid angle range input]
            - I_sc : intensity of scattered field, 
        
        - return_value == "efield" : float
            - [tetalist : teta angles - if solid angle range input]
            - [philist : phi angles - if solid angle range input]
            - E_sc : complex scattered field at each pos.
            
        - return_value == "int_XX" : float
            integrated total intensity over specified solid angle
        
    
    Notes
    -----
    See equation (G1) in
    
    Alaee, R., Rockstuhl, C. & Fernandez-Corbaton, I. *An electromagnetic 
    multipole expansion beyond the long-wavelength approximation.*
    Optics Communications 407, 17–21 (2018)
        
    """
# =============================================================================
#     exception handling
# =============================================================================
    if np.pi < tetamax < 0:
        raise ValueError("`tetamax` out of range, must be in [0, pi]")
    
    if r_probe is not None and return_value in ['int_es', 'int_E0', 'int_Etot']:
        raise ValueError("probing farfield on user-defined positions does not support integration " +
                         "of the intensity since no surface differential can be defined. Use spherical " +
                         "coordinate definition to do the integration.")
    
    p, m, qe, qm, r_p, r_m, r_qe, r_qm, r0 = _check_multipole_list(
        p=p, m=m, qe=qe, qm=qm, r_p=r_p, r_m=r_m, r_qe=r_qe, r_qm=r_qm, r0=r0)
    
        
# =============================================================================
#     preparation
# =============================================================================
    ## --- spherical probe coordinates
    if r_probe is None:
        tetalist = np.ones((int(Nteta), int(Nphi)))*np.linspace(tetamin, tetamax, int(Nteta))[:,None]
        philist = np.ones((int(Nteta), int(Nphi)))*np.linspace(phimin, phimax, int(Nphi), endpoint=False)[None,:]
        xff = (r * np.sin(tetalist) * np.cos(philist)).flatten()
        yff = (r * np.sin(tetalist) * np.sin(philist)).flatten()
        zff = (r * np.cos(tetalist)).flatten()
        _r_probe = np.transpose([xff, yff, zff])
    else:
        _r_probe = r_probe
        
    ## --- spherical integration steps
    dteta = (tetamax-tetamin) / float(Nteta-1)  # endpoint included
    dphi = (phimax-phimin) / float(Nphi)        # endpoint not included
    
    ## --- environment
    k0 = 2*np.pi / wavelength
    n_env = (eps_env**0.5).real
    k = k0*n_env
    
#==============================================================================
#     electric polarization of structure, fundamental field
#==============================================================================        
    Escat = np.zeros(shape=(len(_r_probe), 3), dtype=DTYPE_C)
    
    def eval_prefactors(r0, _r_probe, k, k0):
        n_vec = _r_probe - r0
        R_d = np.linalg.norm(n_vec, axis=1)
        n_vec = n_vec / R_d[:,None]     # unit vector along direction of scattering
        
        phase_factor = np.exp(1j*k*R_d)
        prefactor = k0**2 * phase_factor / R_d
        
        return prefactor, n_vec
    
    
    for _r0, _p in zip(r_p, p):
        prefactor, n0_vec = eval_prefactors(_r0, _r_probe, k, k0)
        Es_p = prefactor[:,None] * np.cross(n0_vec, np.cross((_p)[None,:], n0_vec))
        Escat += Es_p     # sum all electric dipole fields
    
    for _r0, _m in zip(r_m, m):
        prefactor, n0_vec = eval_prefactors(_r0, _r_probe, k, k0)
        Es_m = prefactor[:,None] * np.cross(_m[None,:], n0_vec)
        Escat += Es_m     # sum all magnetic dipole fields
    
    for _r0, _qe in zip(r_qe, qe):
        prefactor, n0_vec = eval_prefactors(_r0, _r_probe, k, k0)
        Es_qe = prefactor[:,None] * 1j * k/6 * np.cross(n0_vec, 
                    np.cross(n0_vec, np.tensordot(_qe, n0_vec, axes=(-1,-1)).T))
        Escat += Es_qe     # sum all electric quadrupole fields
    
    for _r0, _qm in zip(r_qm, qm):
        prefactor, n0_vec = eval_prefactors(_r0, _r_probe, k, k0)
        Es_qm = prefactor[:,None] * 1j * k/6 * np.cross(n0_vec, 
                    np.tensordot(_qm, n0_vec, axes=(-1,-1)).T)
        Escat += Es_qm     # sum all magnetic quadrupole fields

    
    Iscat = np.sum((np.abs(Escat)**2), axis=1)



#==============================================================================
#     return either of scattered field / intensity / map
#==============================================================================
    ## --- total field (no polarizer)
    if r_probe is None:
        out_shape = tetalist.shape
    else:
        out_shape = len(_r_probe)
        
    I_sc  = Iscat.reshape(out_shape)
    
    
    if return_value.lower() in ['map', 'map_i', 'i_map']:
        if r_probe is None:
            return tetalist, philist, I_sc
        else:
            return I_sc
    elif return_value.lower() in ['efield', 'fields', 'field']:
        if r_probe is None:
            return tetalist, philist, Escat
        else:
            return Escat
    else:
        d_solid_surf = r**2 * np.sin(tetalist) * dteta * dphi
        if return_value.lower() in ['int_es', 'int']:
            return np.sum(I_sc * d_solid_surf)
        
        else:
            raise ValueError("Parameter 'return_value' must be one of ['map', 'int_es', 'efield'].")



# =============================================================================
# HIGH-LEVEL (pyGDM2 compatibility)
# =============================================================================
def extinct(sim, field_index, with_toroidal=True, r0=None, eps_dd=0.001, 
            use_generalized_polarizabilities=False, normalization_E0=True,
            long_wavelength_approx=False):
    """extinction cross sections from multipole decomposition
    
    ** ------- FUNCTION STILL UNDER TESTING ------- **
    
    Returns extinction cross sections for electric and magnetic dipole and 
    quadrupole moments of the multipole decomposition.
    
    
    Parameters
    ----------
    sim : :class:`.core.simulation`
        simulation description
    
    field_index : int
        index of evaluated self-consistent field to use for calculation. Can be
        obtained for specific parameter-set using :func:`.tools.get_closest_field_index`
    
    with_toroidal : bool, default: True
        whether to add toroidal moments to electric dipole and quadrupole
    
    r0 : array, default: None
        [x,y,z] position of mulipole decomposition development. 
        If `None`, use structure's center of gravity
    
    eps_dd : float, default: 0.1
        numerical integration step (in nm). Used for e/m quadrupole extinction.
    
    normalization_E0 : bool, default: True
        normalizes sections to max. incident field intensity
        
    long_wavelength_approx : bool, default: False
        if True, use long wavelength approximation
        
        
    Returns
    -------
    sigma_ext_p : float
        electric dipole extinction cross section (in nm^2)
    
    sigma_ext_m : float
        magnetic dipole extinction cross section (in nm^2)
        
    sigma_ext_q : float
        electric quadrupole extinction cross section (in nm^2)
        
    sigma_ext_mq : float
        magnetic quadrupole extinction cross section (in nm^2)
    
    
    Notes
    -----
    For details about the extinction section of multipole moments, see:
    
    Evlyukhin, A. B. et al. *Multipole analysis of light scattering by 
    arbitrary-shaped nanoparticles on a plane surface.*, 
    JOSA B 30, 2589 (2013)
    
    """
# =============================================================================
#     Exception handling
# =============================================================================

# =============================================================================
#     extinction section calculation
# =============================================================================
    from pyGDM2 import linear
    from pyGDM2 import tools
    
    ## by default, use center of gravity for multimode expansion
    geo = sim.struct.geometry
    if r0 is None:
        r0 = np.average(geo, axis=0)
    
    ## get dipole moments
    if not use_generalized_polarizabilities:
        p, p1, m, Qe, Qe1, Qm = multipole_decomposition_exact(
                sim, field_index, r0=r0, 
                long_wavelength_approx=long_wavelength_approx, 
                which_moments=['p', 'p1', 'm', 'qe', 'qe1', 'qm'])
        if not with_toroidal:
            p = p1
            Qe = Qe1
    else:
        if with_toroidal:
            p = eval_generalized_polarizability_p(sim, field_index, which_order='p')
            Qe = eval_generalized_polarizability_qe(sim, field_index, which_order='qe')
        else:
            p = eval_generalized_polarizability_p(sim, field_index, which_order='p1')
            Qe = eval_generalized_polarizability_qe(sim, field_index, which_order='qe1')
        m = eval_generalized_polarizability_m(sim, field_index)
        Qm = eval_generalized_polarizability_qm(sim, field_index)
        
    ## prefactors and incident field
    field_params = tools.get_field_indices(sim)[field_index]
    wavelength   = field_params['wavelength']
    sim.struct.setWavelength(wavelength)
    k0 = 2*np.pi / wavelength
    eps_env = sim.dyads.getEnvironmentIndices(wavelength, geo[:1])[0]  # structure must be fully in one environment zone
    n_env = (eps_env**0.5).real
    
    ## incident field and its gradients at multipole position
    env_dict = sim.dyads.getConfigDictG(wavelength, sim.struct, sim.efield)
    E0 = sim.efield.field_generator(r0[None,:], env_dict, **field_params)
    H0 = sim.efield.field_generator(r0[None,:], env_dict, returnField='H', **field_params)
    
    gradE0, gradH0 = linear.field_gradient(sim, field_index, r0[None,:], 
                                           delta=eps_dd, which_fields=['E0', 'H0'])
    gradE0cj = np.conj(np.array(gradE0)[:,0,3:])
    gradH0cj = np.conj(np.array(gradH0)[:,0,3:])
    
    ## normalization
    if normalization_E0:
        E2in = np.sum(np.abs(E0)**2, axis=1)   # intensity of incident field
    else:
        E2in = 1.0
    prefactor = (4 * np.pi * k0 * 1/n_env * 1/E2in).real
    
    ecs_p = prefactor * (np.sum(np.conjugate(E0)*p)).imag
    ecs_m = prefactor / n_env * (np.sum(np.conjugate(H0)*m)).imag
    ecs_Qe = prefactor / 12. * (np.sum(np.tensordot(gradE0cj + gradE0cj.T, Qe) )).imag
    ecs_Qm = prefactor / n_env / 6 * (np.sum(np.tensordot(gradH0cj.T, Qm) )).imag
    
    return [ecs_p, ecs_m, ecs_Qe, ecs_Qm]
    

def scs(sim, field_index, with_toroidal=True, 
        use_generalized_polarizabilities=False, 
        r0=None, normalization_E0=True,
        long_wavelength_approx=False):
    """total scattering cross section from multipole decomposition
    
    ** ------- FUNCTION STILL UNDER TESTING ------- **
    
    Returns scattering cross sections for electric and magnetic dipole and 
    quadrupole moments of the multipole decomposition.
    
    
    Parameters
    ----------
    sim : :class:`.core.simulation`
        simulation description
    
    field_index : int
        index of evaluated self-consistent field to use for calculation. Can be
        obtained for specific parameter-set using :func:`.tools.get_closest_field_index`
    
    with_toroidal : bool, default: True
        whether to add toroidal moments to electric dipole and quadrupole
            
    use_generalized_polarizabilities : bool, default: False
        if True, use generalized polarizabilities
        (does not require evaluation of `core.scatter` for new incident field 
         configurations, first calculation is more expensive)
    
    r0 : array, default: None
        [x,y,z] position of mulipole decomposition development. 
        If `None`, use structure's center of gravity
    
    normalization_E0 : bool, default: True
        normalizes sections to max. incident field intensity
        
    long_wavelength_approx : bool, default: False
        if True, use long wavelength approximation
    
    
    Returns
    -------
    sigma_scat_p : float
        electric dipole scattering cross section (in nm^2)
    
    sigma_scat_m : float
        magnetic dipole scattering cross section (in nm^2)
        
    sigma_scat_q : float
        electric quadrupole scattering cross section (in nm^2)
        
    sigma_scat_mq : float
        magnetic quadrupole scattering cross section (in nm^2)
    
    
    Notes
    -----
    For details about the exact multipole formalism and scs calculation, see: 
        
    Alaee, R., Rockstuhl, C. & Fernandez-Corbaton, I. *An electromagnetic 
    multipole expansion beyond the long-wavelength approximation.*
    Optics Communications 407, 17–21 (2018)
    
    """
    from pyGDM2 import tools
# =============================================================================
#     Exception handling
# =============================================================================

# =============================================================================
#     scattering section calculation
# =============================================================================
    ## by default, use center of gravity for multimode expansion
    geo = sim.struct.geometry
    if r0 is None:
        r0 = np.average(geo, axis=0)
    
    ## get dipole moments
    if not use_generalized_polarizabilities:
        p, p1, m, Qe, Qe1, Qm = multipole_decomposition_exact(
                sim, field_index, r0=r0,
                which_moments=['p', 'p1', 'm', 'qe', 'qe1', 'qm'], 
                long_wavelength_approx=long_wavelength_approx)
        if not with_toroidal:
            p = p1
            Qe = Qe1
    else:
        if with_toroidal:
            p = eval_generalized_polarizability_p(sim, field_index, which_order='p', long_wavelength_approx=long_wavelength_approx)
            Qe = eval_generalized_polarizability_qe(sim, field_index, which_order='qe', long_wavelength_approx=long_wavelength_approx)
        else:
            p = eval_generalized_polarizability_p(sim, field_index, which_order='p1', long_wavelength_approx=long_wavelength_approx)
            Qe = eval_generalized_polarizability_qe(sim, field_index, which_order='qe1', long_wavelength_approx=long_wavelength_approx)
        m = eval_generalized_polarizability_m(sim, field_index, long_wavelength_approx=long_wavelength_approx)
        Qm = eval_generalized_polarizability_qm(sim, field_index, long_wavelength_approx=long_wavelength_approx)
    
    ## prefactors / illumination fields
    field_params = tools.get_field_indices(sim)[field_index]
    wavelength   = field_params['wavelength']
    sim.struct.setWavelength(wavelength)
    k0 = 2*np.pi / wavelength
    
    eps_env = sim.dyads.getEnvironmentIndices(wavelength, geo[:1])[0]  # assume structure is fully in one environment
    n_env = (eps_env**0.5).real
    k = k0*n_env
    
    ## normalization: incident field intensity at multipole position
    env_dict = sim.dyads.getConfigDictG(wavelength, sim.struct, sim.efield)
    E0 = sim.efield.field_generator(r0[None,:], env_dict, **field_params)
    if normalization_E0:
        E2in = np.sum(np.abs(E0)**2, axis=1)   # intensity of incident field
    else:
        E2in = 1.0
    
    ## factor 100: cm --> m (cgs units)
    sc_factor_dp = 100/12*(k0**4 / E2in).real
    sc_factor_Q = 100/1440*(k0**4 / E2in).real
    
    
    scs_p = sc_factor_dp * np.sum(np.abs(p)**2)
    scs_m = sc_factor_dp * np.sum(np.abs(m)**2)
    scs_Qe = sc_factor_Q * np.sum(np.abs(k*Qe)**2)
    scs_Qm = sc_factor_Q * np.sum(np.abs(k*Qm)**2)
            
    return [scs_p, scs_m, scs_Qe, scs_Qm]



def farfield(sim, field_index, 
             r_probe=None,
             r=100000., 
             tetamin=0, tetamax=np.pi/2., Nteta=10, 
             phimin=0, phimax=2*np.pi, Nphi=36, 
             polarizerangle='none', return_value='map', 
             normalization_E0=False, 
             which_moments=['p', 'm', 'qe', 'qm'],
             long_wavelength_approx=False, r0=None):
    """spatially resolved and polarization-filtered far-field scattering 
    
    For a given incident field, calculate the electro-magnetic field 
    (E-component) in the far-field around the nanostructure 
    (on a sphere of radius `r`).
    
    Propagator for scattering into a substrate contributed by C. Majorel
    
    Pure python implementation.
    
    Parameters
    ----------
    sim : :class:`.core.simulation`
        simulation description
        
    field_index : int
        index of evaluated self-consistent field to use for calculation. Can be
        obtained for specific parameter-set using :func:`.tools.get_closest_field_index`
    
    r_probe : tuple (x,y,z) or list of 3-lists/-tuples. optional. Default: don't use
        defaults to *None*, which means it is not used and a solid angle defined by
        a spherical coordinate range is used instead. If `r_probe` is given, this
        overrides `r`, `tetamin`, `tetamax`, `Nteta`, `Nphi`.
        (list of) coordinate(s) to evaluate farfield on. 
        Format: tuple (x,y,z) or list of 3 lists: [Xmap, Ymap, Zmap] 
        
    r : float, default: 100000.
        radius of integration sphere (distance to coordinate origin in nm)
        
    tetamin, tetamax : float, float; defaults: 0, np.pi/2
        minimum and maximum polar angle in radians 
        (in linear steps from `tetamin` to `tetamax`)
        
    phimin, phimax : float, float; defaults: 0, 2*np.pi
        minimum and maximum azimuth angle in radians, excluding last position
        (in linear steps from `phimin` to `phimax`)
        
    Nteta, Nphi : int, int; defaults: 10, 36
        number of polar and azimuthal angles on sphere to calculate,
        
    polarizerangle : float or 'none', default: 'none'
        optional polarization filter angle **in degrees**(!). If 'none' (default), 
        the total field-intensity is calculated (= no polarization filter)
    
    return_value : str, default: 'map'
        Values to be returned. Either 'map' (default) or 'integrated'.
          - "map" : (default) return spatially resolved farfield intensity at each spherical coordinate (5 lists)
          - "efield" : return spatially resolved E-fields at each spherical coordinate (5 lists)
          - "int_Es" : return the integrated scattered field (as float)
          - "int_E0" : return the integrated fundamental field (as float)
          - "int_Etot" : return the integrated total field (as float)
          
    normalization_E0 : bool, default: False
        has only effect on return_value=="int_Es": Normalizes scattering 
        to peak of incident field intensity inside structure
        
    r0 : array, default: None
        [x,y,z] position of mulipole decomposition development. 
        If `None`, use structure's center of gravity
    
    which_moments : list of str, default: ['p', 'm', 'qe', 'qm']
        which multipole moments to use for farfield calculations. Supported:
            - 'p': electric dipole (full)
            - 'm': magnetic dipole
            - 'qe': electric quadrupole (full)
            - 'qm': magnetic quadrupole
            - 'p1': electric dipole (first order)
            - 'pt': toroidal dipole
            - 'qe1': electric quadrupole (first order)
            - 'qet': toroidal quadrupole
            
    long_wavelength_approx : bool, default: False
        if True, use long wavelength approximation
    
    Returns
    -------
    using `r_probe` for position definition:
        3 lists of 6-tuples (x,y,z, Ex,Ey,Ez), complex : 
            - scattered Efield or E-intensity
            - total Efield (inlcuding fundamental field) or total E intensity
            - fundamental Efield (incident field) or E0 intensity
        
    if solid angle is defined via spherical coordinate range:
        - return_value == "map" : 5 arrays of shape (Nteta, Nphi) : 
            - [tetalist : teta angles - if solid angle range input]
            - [philist : phi angles - if solid angle range input]
            - I_sc : intensity of scattered field, 
            - I_tot : intensity of total field (I_tot=|E_sc+E_0|^2), 
            - I0 : intensity of incident field
        
        - return_value == "efield" : float
            - [tetalist : teta angles - if solid angle range input]
            - [philist : phi angles - if solid angle range input]
            - E_sc : complex scattered field at each pos.
            - E_tot : complex total field at each pos. (E_sc+E0)
            - E0 : complex incident field at each pos.
            
        - return_value == "int_XX" : float
            integrated total intensity over specified solid angle
        
    Notes
    -----
    See equation (G1) in
    
    Alaee, R., Rockstuhl, C. & Fernandez-Corbaton, I. *An electromagnetic 
    multipole expansion beyond the long-wavelength approximation.*
    Optics Communications 407, 17–21 (2018)
        
    """
# =============================================================================
#     exception handling
# =============================================================================
    if sim.E is None: 
        raise ValueError("Error: Scattering field inside the structure not yet evaluated. Run `core.scatter` simulation first.")
    
    if str(polarizerangle).lower() == 'none':
        polarizer = 0
    else:
        polarizer = polarizerangle * np.pi/180.
    
    if return_value.lower().startswith('int') and (0 > tetamin >= tetamax):
        raise ValueError("`tetamin` out of range, must be in [0, tetamax] for integration.")
    
    if return_value.lower().startswith('int') and (np.pi < tetamax < 0):
        raise ValueError("`tetamax` out of range, must be in [0, pi] for integration.")
    
    if phimin < 0:
        warnings.warn("phi is supposed to be a positive number, use at own risk.")
    if phimin > phimax:
        raise ValueError("`phimin` out of range, must be <= phimax")
    
    if return_value.lower().startswith('int') and (0 > phimax > 2*np.pi):
        raise ValueError("`phimax` out of range, must be in [0, 2*pi] for integration.")
    
    if r_probe is not None and return_value in ['int_es', 'int_E0', 'int_Etot']:
        raise ValueError("probing farfield on user-defined positions does not support integration " +
                         "of the intensity since no surface differential can be defined. Use spherical " +
                         "coordinate definition to do the integration.")
    
    which_moments = [wm.lower() for wm in which_moments]
    if ('p' in which_moments and 'p1' in which_moments) or ('p' in which_moments and 'pt' in which_moments):
        raise Exception("Please use one single electric dipole moment for farfield propagation.")
    if ('qe' in which_moments and 'qe1' in which_moments) or ('qe' in which_moments and 'qet' in which_moments):
        raise Exception("Please use one single electric quadrupole moment for farfield propagation.")
        
# =============================================================================
#     preparation
# =============================================================================
    from pyGDM2 import tools
    ## --- spherical probe coordinates
    if r_probe is None:
        tetalist = np.ones((int(Nteta), int(Nphi)))*np.linspace(tetamin, tetamax, int(Nteta))[:,None]
        philist = np.ones((int(Nteta), int(Nphi)))*np.linspace(phimin, phimax, int(Nphi), endpoint=False)[None,:]
        xff = (r * np.sin(tetalist) * np.cos(philist)).flatten()
        yff = (r * np.sin(tetalist) * np.sin(philist)).flatten()
        zff = (r * np.cos(tetalist)).flatten()
        _r_probe = np.transpose([xff, yff, zff])
    else:
        _r_probe = r_probe
        
    ## --- spherical integration steps
    dteta = (tetamax-tetamin) / float(Nteta-1)  # endpoint included
    dphi = (phimax-phimin) / float(Nphi)        # endpoint not included
    
    ## --- incident field config
    field_params    = tools.get_field_indices(sim)[field_index]
    wavelength      = field_params['wavelength']
    
    ## --- structure
    geo = sim.struct.geometry
    if r0 is None:
        r0 = np.average(geo, axis=0)
    n_vec = _r_probe - r0
    R_d = np.linalg.norm(n_vec, axis=1)
    n0_vec = n_vec / R_d[:,None]     # unit vector for direction of scattering
    
    ## --- environment
    sim.struct.setWavelength(wavelength)
    k0 = 2*np.pi / wavelength
    eps_env = sim.dyads.getEnvironmentIndices(wavelength, geo[:1])[0]  # assume structure is fully in one environment
    n_env = (eps_env**0.5).real
    k = k0*n_env
    
#==============================================================================
#     electric polarization of structure, fundamental field
#==============================================================================        
    ## --- fundamental field - use dummy structure with 
    env_dict = sim.dyads.getConfigDictG(wavelength, sim.struct, sim.efield)
    if return_value == 'int_Es':
        from pyGDM2 import fields
        E0 = fields.nullfield(_r_probe, env_dict, wavelength, returnField='E')
    else:
        E0 = sim.efield.field_generator(_r_probe, env_dict, **field_params)
        
    
    p, p1, pt, m, Qe, Qe1, Qet, Qm = multipole_decomposition_exact(
                sim, field_index, r0=r0,
                which_moments=['p', 'p1', 'pt', 'm', 'qe', 'qe1', 'qet', 'qm'], 
                long_wavelength_approx=long_wavelength_approx)
    
    phase_factor = np.exp(1j*k*R_d)
    prefactor = k0**2 * phase_factor / (R_d)
    
    Es_p1 = prefactor[:,None] * np.cross(n0_vec, np.cross((p1)[None,:], n0_vec))
    Es_pt = prefactor[:,None] * np.cross(n0_vec, np.cross((pt)[None,:], n0_vec))    
    Es_p = Es_p1 + Es_pt
    Es_m = prefactor[:,None] * np.cross(m[None,:], n0_vec)
    
    Es_qe1 = prefactor[:,None] * 1j * k/6 * np.cross(n0_vec, 
                 np.cross(n0_vec, np.tensordot(Qe1, n0_vec, axes=(-1,-1)).T))
    Es_qet = prefactor[:,None] * 1j * k/6 * np.cross(n0_vec, 
                 np.cross(n0_vec, np.tensordot(Qet, n0_vec, axes=(-1,-1)).T))
    Es_qe = Es_qe1 + Es_qet
    Es_qm = prefactor[:,None] * 1j * k/6 * np.cross(n0_vec, 
                                    np.tensordot(Qm, n0_vec, axes=(-1,-1)).T)

    ## sum up all multipole fields
    Escat = np.zeros(shape=(len(_r_probe), 3), dtype=sim.efield.dtypec)
    if ('p1' in which_moments and 'pt' in which_moments) or 'p' in which_moments:
        Escat += Es_p
    elif 'p1' in which_moments:
        Escat += Es_p1
    elif 'pt' in which_moments:
        Escat += Es_pt
    if 'm' in which_moments:
        Escat += Es_m
    if ('qe1' in which_moments and 'qet' in which_moments) or 'qe' in which_moments:
        Escat += Es_qe
    elif 'qe1' in which_moments:
        Escat += Es_qe1
    elif 'qet' in which_moments:
        Escat += Es_qet
    if 'qm' in which_moments:
        Escat += Es_qm
    
    
    Iscat = np.sum((np.abs(Escat)**2), axis=1)


#==============================================================================
#    calc. fields through optional polarization filter
#==============================================================================
    if str(polarizerangle).lower() != 'none':
        ## --- scattered E-field parallel and perpendicular to scattering plane
        Es_par  = ( Escat.T[0] * np.cos(tetalist.flatten()) * np.cos(philist.flatten()) + 
                    Escat.T[1] * np.sin(philist.flatten()) * np.cos(tetalist.flatten()) - 
                    Escat.T[2] * np.sin(tetalist.flatten()) )
        Es_perp = ( Escat.T[0] * np.sin(philist.flatten()) - Escat.T[1] * np.cos(philist.flatten()) )
        ## --- scattered E-field parallel to polarizer
        Es_pol  = ( Es_par * np.cos(polarizer - philist.flatten()) - 
                    Es_perp * np.sin(polarizer - philist.flatten()) )
        
        ## --- fundamental E-field parallel and perpendicular to scattering plane
        E0_par  = ( E0.T[0] * np.cos(tetalist.flatten()) * np.cos(philist.flatten()) + 
                    E0.T[1] * np.sin(philist.flatten()) * np.cos(tetalist.flatten()) - 
                    E0.T[2] * np.sin(tetalist.flatten()) )
        E0_perp = ( E0.T[0] * np.sin(philist.flatten()) - E0.T[1] * np.cos(philist.flatten()) )
        ## --- fundamental E-field parallel to polarizer
        E0_pol  = ( E0_par * np.cos(polarizer - philist.flatten()) - 
                    E0_perp * np.sin(polarizer - philist.flatten()) )

#==============================================================================
#     Intensities with and without fundamental field / polarizer
#==============================================================================
    ## --- total field (no polarizer)
    if r_probe is None:
        out_shape = tetalist.shape
    else:
        out_shape = len(_r_probe)
        
    I_sc  = Iscat.reshape(out_shape)
    I0    = np.sum((np.abs(E0)**2), axis=1).reshape(out_shape)
    I_tot = np.sum((np.abs(E0 + Escat)**2), axis=1).reshape(out_shape)
    
    ## --- optionally: with polarizer
    if str(polarizerangle).lower() != 'none':
        I_sc  = (np.abs(Es_pol)**2).reshape(out_shape)
        I0    = (np.abs(E0_pol)**2).reshape(out_shape)
        I_tot = (np.abs(Es_pol + E0_pol)**2).reshape(out_shape)
    
    
    if return_value.lower() == 'map':
        if r_probe is None:
            return tetalist, philist, I_sc, I_tot, I0
        else:
            return I_sc, I_tot, I0
    elif return_value.lower() in ['efield', 'fields', 'field']:
        if r_probe is None:
            return tetalist, philist, Escat, Escat + E0, E0
        else:
            return Escat, Escat + E0, E0
    else:
        d_solid_surf = r**2 * np.sin(tetalist) * dteta * dphi
        if return_value.lower() == 'int_es':
            if normalization_E0:
                env_dict = sim.dyads.getConfigDictG(wavelength, sim.struct, sim.efield)
                E0 = sim.efield.field_generator(sim.struct.geometry, 
                                                env_dict, **field_params)
                I0_norm = np.sum(np.abs(E0)**2, axis=1).max()
            else:
                I0_norm = 1
            
            return np.sum(I_sc * d_solid_surf) / I0_norm
        
        elif return_value.lower() == 'int_e0':
            return np.sum(I0 * d_solid_surf)
        
        elif return_value.lower() == 'int_etot':
            return np.sum(I_tot * d_solid_surf)
        
        else:
            raise ValueError("Parameter 'return_value' must be one of ['map', 'int_es', 'int_e0', 'int_etot'].")









# =============================================================================
# multipole modes
# =============================================================================
def density_of_multipolar_modes(sim, field_index=None, wavelength=None, return_mode_energy=True, 
                                which_moments=['p', 'm', 'qe', 'qm'], 
                                method='lu', long_wavelength_approx=False):
    """Calculate total density of multipole modes via generalized polarizability
    
    If not yet calculated, will invoke generalize pola. calculation, thus it may
    require some computation time on the first call.
    

    Parameters
    ----------
    sim : :class:`.core.simulation`
        simulation description
    
    field_index : int
        index of evaluated self-consistent field to use for calculation. Can be
        obtained for specific parameter-set using :func:`.tools.get_closest_field_index`
    
    energy_density : bool, optional
        whether to return the total mode energy (True) or the accumulated 
        multipole amplitude norm (False). The default is True.
    
    wavelength : float, default: None
        Optional wavelength (alternative to `field_index`) at which to 
        calculate susceptibility matrix (in nm). 
        Either `field_index` or `wavelength` must be given.
    
    which_moments : list of str, default: ['p', 'm', 'qe', 'qm']
        which multipole moments to calculate and return. supported dipole moments: 
            - 'p': electric dipole (full)
            - 'm': magnetic dipole
            - 'qe': electric quadrupole (full)
            - 'qm': magnetic quadrupole
            - 'p1': electric dipole (first order)
            - 'pt': toroidal dipole
            - 'qe1': electric quadrupole (first order)
            - 'qet': toroidal quadrupole
    
    method : string, default: "lu"
        inversion method if generalized pola. not yet calculated. 
        One of ["lu", "numpyinv", "scipyinv", "cupy", "cuda"]
    
    long_wavelength_approx : bool, default: False
        if True, use long wavelength approximation
    

    Returns
    -------
    out_list : list of float
        total density of modes for each demanded order (= sum of tensor norms 
        of all generalized polarizabilities for each multipole)
    
    
    Notes
    -----
    For details , see: 
        
    *PAPER SUBMITTED*

    """
    # =============================================================================
    #     Exception handling
    # =============================================================================
    if field_index is None and wavelength is None:
        raise Exception("Either `field_index` or `wavelength` must be given!")
        
    if field_index is not None and wavelength is not None:
        warnings.warn("`field_index` AND `wavelength` are given! Ignoring `wavelength`.")
        
    
    # =============================================================================
    # preparation
    # =============================================================================
    if field_index is not None:
        from pyGDM2 import tools
        wavelength = tools.get_field_indices(sim)[field_index]['wavelength']
    
    which_moments = [wm.lower() for wm in which_moments]
    for which_moment in which_moments:
        _test_availability_generalized_polarizability(sim, which_moment, 
                                                      wavelength, method=method, 
                                                      long_wavelength_approx=long_wavelength_approx)
    exp_order = 2 if return_mode_energy else 1
    
    out_list = []
    for wm in which_moments:
        if 'p' == wm:
            out_list.append((np.sum(np.linalg.norm(sim.K_P_E[wavelength], axis=(1,2))) + 
                            np.sum(np.linalg.norm(sim.K_T_E[wavelength], axis=(1,2))))**exp_order)
        if 'p1' == wm:
            out_list.append(np.sum(np.linalg.norm(sim.K_P_E[wavelength], axis=(1,2)))**exp_order)
        if 'pt' == wm:
            out_list.append(np.sum(np.linalg.norm(sim.K_T_E[wavelength], axis=(1,2)))**exp_order)
        if 'qe' == wm:
            out_list.append((np.sum(np.linalg.norm(sim.K_QE_E[wavelength], axis=(1,2))) +
                            np.sum(np.linalg.norm(sim.K_QT_E[wavelength], axis=(1,2))))**exp_order)
        if 'qe1' == wm:
            out_list.append(np.sum(np.linalg.norm(sim.K_QE_E[wavelength], axis=(1,2)))**exp_order)
        if 'qt' == wm:
            out_list.append(np.sum(np.linalg.norm(sim.K_QT_E[wavelength], axis=(1,2)))**exp_order)
        if 'm' == wm:
            out_list.append(np.sum(np.linalg.norm(sim.K_M_E[wavelength], axis=(1,2)))**exp_order)
        if 'qm' == wm:
            out_list.append(np.sum(np.linalg.norm(sim.K_QM_E[wavelength], axis=(1,2)))**exp_order)
    
    return out_list




